import os
# work with proc_per_node 2
# os.environ["CUDA_VISIBLE_DEVICES"] = "0,2"
import torch
import contextlib
from torch.nn.parallel import DistributedDataParallel as DDP
import torch.distributed as dist
import pkg_resources as pkg
from contextlib import contextmanager
from torch.cuda import amp
from datetime import datetime
from tqdm import tqdm
import math
from copy import deepcopy
import torch.nn as nn
from aicmder.utils import colorstr
from pathlib import Path
torch.backends.cudnn.benchmark = True

LOCAL_RANK = int(os.getenv('LOCAL_RANK', -1))  # https://pytorch.org/docs/stable/elastic/run.html
RANK = int(os.getenv('RANK', -1))
WORLD_SIZE = int(os.getenv('WORLD_SIZE', 1))


def smart_optimizer(model, name='Adam', lr=0.001, momentum=0.9, decay=1e-5):
    # YOLOv5 3-param group optimizer: 0) weights with decay, 1) weights no decay, 2) biases no decay
    g = [], [], []  # optimizer parameter groups
    bn = tuple(v for k, v in nn.__dict__.items() if 'Norm' in k)  # normalization layers, i.e. BatchNorm2d()
    for idx, v in enumerate(model.modules()):
        # weight and bias, here add bias
        if hasattr(v, 'bias') and isinstance(v.bias, nn.Parameter):  # bias (no decay)
            g[2].append(v.bias)

        # add weight
        if isinstance(v, bn):  # weight (no decay)
            g[1].append(v.weight)
        elif hasattr(v, 'weight') and isinstance(v.weight, nn.Parameter):  # weight (with decay)
            g[0].append(v.weight)

    if name == 'Adam':
        optimizer = torch.optim.Adam(g[2], lr=lr, betas=(momentum, 0.999))  # adjust beta1 to momentum
    elif name == 'AdamW':
        optimizer = torch.optim.AdamW(g[2], lr=lr, betas=(momentum, 0.999), weight_decay=0.0)
    elif name == 'RMSProp':
        optimizer = torch.optim.RMSprop(g[2], lr=lr, momentum=momentum)
    elif name == 'SGD':
        optimizer = torch.optim.SGD(g[2], lr=lr, momentum=momentum, nesterov=True)
    else:
        raise NotImplementedError(f'Optimizer {name} not implemented.')

    optimizer.add_param_group({'params': g[0], 'weight_decay': decay})  # add g0 with weight_decay
    optimizer.add_param_group({'params': g[1], 'weight_decay': 0.0})  # add g1 (BatchNorm2d weights)
    print(f"{colorstr('optimizer:')} {type(optimizer).__name__}(lr={lr}) with parameter groups "
                f"{len(g[1])} weight(decay=0.0), {len(g[0])} weight(decay={decay}), {len(g[2])} bias")
    return optimizer

class EarlyStopping:
    # YOLOv5 simple early stopper
    def __init__(self, evaluate_func, patience=30):
        self.best_fitness = None  # i.e. mAP
        self.best_epoch = 0
        self.patience = patience or float('inf')  # epochs to wait after fitness stops improving to stop
        self.possible_stop = False  # possible stop may occur next epoch
        self.evaluate_func = evaluate_func

    def __call__(self, epoch, fitness):
        if self.best_fitness is None:
            self.best_fitness = fitness
            
        if self.evaluate_func(fitness, self.best_fitness):  # >= 0 to allow for early zero-fitness stage of training
            self.best_epoch = epoch
            self.best_fitness = fitness
        delta = epoch - self.best_epoch  # epochs without improvement
        self.possible_stop = delta >= (self.patience - 1)  # possible stop may occur next epoch
        stop = delta >= self.patience  # stop training if patience exceeded
        if stop:
            print(f'Stopping training early as no improvement observed in last {self.patience} epochs. '
                        f'Best results observed at epoch {self.best_epoch}, best model saved as best.pt.\n'
                        f'To update EarlyStopping(patience={self.patience}) pass a new patience value, '
                        f'i.e. `python train.py --patience 300` or use `--patience 0` to disable EarlyStopping.')
        return stop



def copy_attr(a, b, include=(), exclude=()):
    # Copy attributes from b to a, options to only include [...] and to exclude [...]
    for k, v in b.__dict__.items():
        if (len(include) and k not in include) or k.startswith('_') or k in exclude:
            continue
        else:
            setattr(a, k, v)


def is_parallel(model):
    # Returns True if model is of type DP or DDP
    return type(model) in (nn.parallel.DataParallel, nn.parallel.DistributedDataParallel)


def de_parallel(model):
    # De-parallelize a model: returns single-GPU model if model is of type DP or DDP
    return model.module if is_parallel(model) else model


@contextmanager
def torch_distributed_zero_first(local_rank: int):
    """
    Decorator to make all processes in distributed training wait for each local_master to do something.
    use LOCAL_RANK here for creating dataset and distributed.DistributedSampler to sample dataset
    """
    if local_rank not in [-1, 0]:
        dist.barrier(device_ids=[local_rank])
    yield
    if local_rank == 0:
        dist.barrier(device_ids=[0])


def check_version(current='0.0.0', minimum='0.0.0', name='version ', pinned=False, hard=False, verbose=False):
    # Check version vs. required version
    current, minimum = (pkg.parse_version(x) for x in (current, minimum))
    result = (current == minimum) if pinned else (current >= minimum)  # bool
    # s = f'{name}{minimum} required by YOLOv5, but {name}{current} is currently installed'  # string
    if hard:
        assert result  # assert min requirements met
    # if verbose and not result:
    #     LOGGER.warning(s)
    return result


def smart_inference_mode(torch_1_9=check_version(torch.__version__, '1.9.0')):
    # Applies torch.inference_mode() decorator if torch>=1.9.0 else torch.no_grad() decorator
    def decorate(fn):
        return (torch.inference_mode if torch_1_9 else torch.no_grad)()(fn)

    return decorate


class DPModel(contextlib.ContextDecorator):
    def __init__(self, model):
        # self.devices = devices
        # os.environ["CUDA_VISIBLE_DEVICES"] = self.devices
        self.gpu_count = torch.cuda.device_count()
        self.model = model

    def __enter__(self):
        if self.gpu_count > 1:
            return torch.nn.DataParallel(self.model)
        else:
            return self.model

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def smart_DDP(model):
    # Model DDP creation with checks
    assert not check_version(torch.__version__, '1.12.0', pinned=True), \
        'torch==1.12.0 torchvision==0.13.0 DDP training is not supported due to a known issue. ' \
        'Please upgrade or downgrade torch to use DDP. See https://github.com/ultralytics/yolov5/issues/8395'
    if check_version(torch.__version__, '1.11.0'):
        return DDP(model, device_ids=[LOCAL_RANK], output_device=LOCAL_RANK, static_graph=True)
    else:
        return DDP(model, device_ids=[LOCAL_RANK], output_device=LOCAL_RANK)


class DDPModel(contextlib.ContextDecorator):
    def __init__(self, model, devices=None, auto_clean=True):
        self.devices = devices
        # os.environ["CUDA_VISIBLE_DEVICES"] = self.devices
        # self.gpu_count = torch.cuda.device_count()
        self.model = model
        self.auto_clean = auto_clean

    def __enter__(self):
        if LOCAL_RANK != -1:
            assert torch.cuda.device_count() > LOCAL_RANK, 'insufficient CUDA devices for DDP command'
            torch.cuda.set_device(LOCAL_RANK)
            device = torch.device('cuda', LOCAL_RANK)
            dist.init_process_group(backend="nccl" if dist.is_nccl_available() else "gloo")
            self.model.to(device)
            cuda = device.type != 'cpu'
            if cuda and RANK != -1:
                return smart_DDP(self.model), device
        self.model.to(self.devices)
        return self.model, self.devices

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def cleanup(self):
        if WORLD_SIZE > 1 and RANK == 0 and self.auto_clean:
            dist.destroy_process_group()


def check_amp(model):
    # Check PyTorch Automatic Mixed Precision (AMP) functionality. Return True on correct operation

    def amp_allclose(model):
        model = model.eval()

        # All close FP32 vs AMP results
        first_parameter = next(model.parameters())
        input_shape = first_parameter.size()
        model_input = torch.zeros(input_shape, device=next(model.parameters()).device).contiguous()  # input
        a = model(model_input)  # FP32 inference

        with amp.autocast(True):
            b = model(model_input)  # AMP inference
        return a.shape == b.shape and torch.allclose(a, b, atol=0.1)  # close to 10% absolute tolerance

    # prefix = colorstr('AMP: ')
    device = next(model.parameters()).device  # get model device
    if device.type == 'cpu':
        return False  # AMP disabled on CPU
    # f = ROOT / 'data' / 'images' / 'bus.jpg'  # image to check
    # im = f if f.exists() else 'https://ultralytics.com/images/bus.jpg' if check_online() else np.ones((640, 640, 3))
    try:
        assert amp_allclose(model)
        # LOGGER.info(emojis(f'{prefix}checks passed ✅'))
        return True
    except Exception as e:
        print("check_amp assert:", e)
        # help_url = 'https://github.com/ultralytics/yolov5/issues/7908'
        # LOGGER.warning(emojis(f'{prefix}checks failed ❌, disabling Automatic Mixed Precision. See {help_url}'))
        return False


class ModelEMA:
    """ Updated Exponential Moving Average (EMA) from https://github.com/rwightman/pytorch-image-models
    Keeps a moving average of everything in the model state_dict (parameters and buffers)
    For EMA details see https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage
    """

    def __init__(self, model, decay=0.9999, tau=2000, updates=0):
        # Create EMA
        self.ema = deepcopy(de_parallel(model)).eval()  # FP32 EMA
        self.updates = updates  # number of EMA updates
        self.decay = lambda x: decay * (1 - math.exp(-x / tau))  # decay exponential ramp (to help early epochs)
        for p in self.ema.parameters():
            p.requires_grad_(False)

    def update(self, model):
        # Update EMA parameters
        self.updates += 1
        d = self.decay(self.updates)

        msd = de_parallel(model).state_dict()  # model state_dict
        for k, v in self.ema.state_dict().items():
            if v.dtype.is_floating_point:  # true for FP16 and FP32
                v *= d
                v += (1 - d) * msd[k].detach()
        # assert v.dtype == msd[k].dtype == torch.float32, f'{k}: EMA {v.dtype} and model {msd[k].dtype} must be FP32'

    def update_attr(self, model, include=(), exclude=('process_group', 'reducer')):
        # Update EMA attributes
        copy_attr(self.ema, model, include, exclude)


def smart_resume(ckpt, optimizer, ema=None, epochs=300, resume=True):
    # Resume training from a partially trained checkpoint
    best_fitness = None
    start_epoch = ckpt['epoch'] + 1
    if ckpt['optimizer'] is not None:
        optimizer.load_state_dict(ckpt['optimizer'])  # optimizer
        best_fitness = ckpt['best_fitness']
    if ema and ckpt.get('ema'):
        ema.ema.load_state_dict(ckpt['ema'].float().state_dict())  # EMA
        ema.updates = ckpt['updates']
    if resume:
        assert start_epoch > 0, f'training to {epochs} epochs is finished, nothing to resume.\n' \
                                f"Start a new training without --resume, i.e. 'python train.py --weights'"
    if epochs < start_epoch:
        epochs += ckpt['epoch']  # finetune additional epochs
    return best_fitness, start_epoch, epochs


def load_model(weight, is_val=True):
    ckpt = torch.load(weight, map_location='cpu')  # load
    if is_val:
        model = (ckpt.get('ema') or ckpt['model']).float()  # FP32 model
    else:
        model = (ckpt.get('model') or ckpt['ema']).float()  # FP32 model
    # model.load_state_dict(csd, strict=False)  # load
    return model, ckpt


class Trainable:

    def __init__(self, bs, nbs=None, gradient_clip_val=10.0) -> None:
        self.pbar = None
        self.ema = None
        self.last_opt_step = -1
        self.bs = bs
        self.train_size = None
        self.gradient_clip_val = gradient_clip_val
        if nbs is not None:
            self.accumulate = max(round(nbs / bs), 1)
        else:
            self.accumulate = 1


    def on_pretrain_routine_start(self, model, verbose=True):
        from aicmder.benchmark.torch_bench import model_info
        if RANK in {-1, 0}:
            model_info(model, verbose=verbose)

    def on_train_start(self, model, device, evaluate_func=None, patience=30, use_ema=False):
        cuda = device.type != 'cpu'
        if cuda and RANK == -1 and torch.cuda.device_count() > 1:
            with DPModel(model) as m:
                model = m.to(device)

        with DDPModel(model, auto_clean=False, devices=device) as ddp_model:
            model, device = ddp_model
            amp = check_amp(model)
            scaler = torch.cuda.amp.GradScaler(enabled=amp)   
            self.ema = None
            if use_ema:     
                try:
                    # EMA
                    self.ema = ModelEMA(model) if RANK in {-1, 0} else None
                except:
                    pass

            if evaluate_func is not None:
                self.stopper = EarlyStopping(evaluate_func, patience=patience)
            return model, device, amp, scaler, self.ema

    def on_train_epoch_start(self, model, epoch, loader, verbose=True):
        model.train()
        if RANK != -1:
            loader.sampler.set_epoch(epoch)

        pbar = enumerate(loader)
        if verbose:
            if RANK in {-1, 0}:
                nb = len(loader)  # number of batches
                # progress bar
                pbar = tqdm(pbar, total=nb,
                            bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}')
        self.pbar = pbar
        if self.train_size is None:
            self.train_size = len(loader.dataset)
        return pbar, self.train_size

    def on_train_batch_end(self, model, scaler, optimizer, loss, desc='', ni=None):
        # old--------
        # loss.backward()
        # old--------

        ############### Backward first
        scaler.scale(loss).backward()

        if ni is None or (ni is not None and ni - self.last_opt_step >= self.accumulate):
            # Only update weights every other accumulate iterations
            scaler.unscale_(optimizer)  # unscale gradients
            torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=self.gradient_clip_val)  # clip gradients
            scaler.step(optimizer)  # optimizer.step
            scaler.update()
            # Set gradients to None (e.g., model.zero_grad(set_to_none=True) ) before the optimizer updates the weights
            optimizer.zero_grad(set_to_none=True)
            if self.ema is not None:
                self.ema.update(model)
            if ni is not None:
                self.last_opt_step = ni
            
        # old--------
        # optimizer.step()
        # old--------
        # optimizer.zero_grad()

        if RANK in {-1, 0}:
            if self.pbar is not None:
                self.pbar.set_description(desc)


    def on_train_epoch_end(self, epoch, fi, scheduler=None):
        if scheduler is not None:
            scheduler.step()
        stop = False
        if hasattr(self, 'stopper'):
            stop = self.stopper(epoch=epoch, fitness=fi)  # early stop check
            # EarlyStopping
            if RANK != -1:  # if DDP training
                broadcast_list = [stop if RANK == 0 else None]
                dist.broadcast_object_list(broadcast_list, 0)  # broadcast 'stop' to all ranks
                if RANK != 0:
                    stop = broadcast_list[0]
        return stop
    

    def on_train_end(self):
        if WORLD_SIZE > 1 and RANK == 0:
            dist.destroy_process_group()

    def on_model_save(self, model, epoch, optimizer, best_fitness, dirname, save_state_dict_only=False):
        if RANK in {-1, 0}:
            if save_state_dict_only: 
                state_dict = model.state_dict()
            else:
                state_dict = deepcopy(de_parallel(model)).half()
            
            ckpt = {
                'epoch': epoch,
                'best_fitness': best_fitness,
                # 'model': deepcopy(de_parallel(model)).half(),
                'model': state_dict,
                'optimizer': optimizer.state_dict(),
                # 'opt': vars(opt),
                # 'wandb_id': None,  # loggers.wandb.wandb_run.id if loggers.wandb else None,
                'date': datetime.now().isoformat()}
            if self.ema is not None:
                ckpt['ema'] = deepcopy(self.ema.ema).half()
                ckpt['updates'] = self.ema.updates

            if os.path.isdir(dirname):
                save_dir = Path(dirname)
                w = save_dir / 'train'  # weights dir
                w.mkdir(parents=True, exist_ok=True)  # make dir
                best_pt = w / 'best.pt'
            else:
                best_pt = dirname
                save_dir = Path(os.path.dirname(best_pt))
                save_dir.mkdir(parents=True, exist_ok=True)  # make dir
            torch.save(ckpt, best_pt)
            del ckpt


# https://pytorch.org/tutorials/beginner/dist_overview.html
# python -m torch.distributed.run --nproc_per_node 2 /home/faith/old_f/aicmder/aicmder/torch/dist_model.py
# 2 0 0
# True
# 2 1 1
# True

def evaluate_func(fitness, best_fitness):
    best_train_acc, best_train_loss = best_fitness
    train_acc, train_loss = fitness
    if train_acc > best_train_acc or (train_acc == best_train_acc and best_train_loss is not None and train_loss < best_train_loss):
        return True
    else: 
        return False
     

if __name__ == "__main__":
    
    stopper = EarlyStopping(evaluate_func)
    stopper(10, (70, 100))
    stopper(60, (90, 100))
    stopper(80, (90, 80))
    stopper(380, (90, 70))
    print(stopper.best_epoch, stopper.best_fitness)


# src: https://github.com/OpenNMT/OpenNMT-py/blob/master/onmt/modules/multi_headed_attn.py
""" Multi-Head Attention module """
import math
import torch
import torch.nn as nn

from aicmder.torch.transformer.misc import generate_relative_positions_matrix,\
                            relative_matmul
# from onmt.utils.misc import aeq

from torch.nn.init import xavier_uniform_


ATTEN_TYPE_SELF = "self"
ATTEN_TYPE_HYGRA = "hydra"


class MultiHeadedAttention(nn.Module):
    """Multi-Head Attention module from "Attention is All You Need"
    :cite:`DBLP:journals/corr/VaswaniSPUJGKP17`.
    Similar to standard `dot` attention but uses
    multiple attention distributions simulataneously
    to select relevant items.
    .. mermaid::
       graph BT
          A[key]
          B[value]
          C[query]
          O[output]
          subgraph Attn
            D[Attn 1]
            E[Attn 2]
            F[Attn N]
          end
          A --> D
          C --> D
          A --> E
          C --> E
          A --> F
          C --> F
          D --> O
          E --> O
          F --> O
          B --> O
    Also includes several additional tricks.
    Args:
       head_count (int): number of parallel heads
       model_dim (int): the dimension of keys/values/queries,
           must be divisible by head_count
       dropout (float): dropout parameter
    """

    def __init__(self, head_count, model_dim, dropout=0.1,
                 max_relative_positions=0, use_neg_dist=True):
        assert model_dim % head_count == 0
        self.dim_per_head = model_dim // head_count
        self.model_dim = model_dim

        super(MultiHeadedAttention, self).__init__()
        self.head_count = head_count

        self.linear_keys = nn.Linear(model_dim,
                                     head_count * self.dim_per_head)
        self.linear_values = nn.Linear(model_dim,
                                       head_count * self.dim_per_head)
        self.linear_query = nn.Linear(model_dim,
                                      head_count * self.dim_per_head)
        self.softmax = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)
        self.final_linear = nn.Linear(model_dim, model_dim)

        self.max_relative_positions = max_relative_positions

        self.use_neg_dist = use_neg_dist
    
        if max_relative_positions > 0:
            vocab_size = max_relative_positions * 2 + 1 \
                if self.use_neg_dist else max_relative_positions + 1
            self.relative_positions_embeddings = nn.Embedding(
                vocab_size, self.dim_per_head)
        self._reset_parameters()

    def _reset_module_parameters(self, module):
        for p in module.parameters():
            if p.dim() > 1:
                xavier_uniform_(p)

    def _reset_parameters(self):
        self._reset_module_parameters(self.linear_keys)
        self._reset_module_parameters(self.linear_values)
        self._reset_module_parameters(self.linear_query)


    def forward(self, key, value, query, mask=None,
                layer_cache=None, attn_type=None):
        """
        Compute the context vector and the attention vectors.
        Args:
           key (FloatTensor): set of `key_len`
               key vectors ``(batch, key_len, dim)``
           value (FloatTensor): set of `key_len`
               value vectors ``(batch, key_len, dim)``
           query (FloatTensor): set of `query_len`
               query vectors  ``(batch, query_len, dim)``
           mask: binary mask 1/0 indicating which keys have
               zero / non-zero attention ``(batch, query_len, key_len)``
        Returns:
           (FloatTensor, FloatTensor):
           * output context vectors ``(batch, query_len, dim)``
           * Attention vector in heads ``(batch, head, query_len, key_len)``.
        """

        # CHECKS
        # batch, k_len, d = key.size()
        # batch_, k_len_, d_ = value.size()
        # aeq(batch, batch_)
        # aeq(k_len, k_len_)
        # aeq(d, d_)
        # batch_, q_len, d_ = query.size()
        # aeq(batch, batch_)
        # aeq(d, d_)
        # aeq(self.model_dim % 8, 0)
        # if mask is not None:
        #    batch_, q_len_, k_len_ = mask.size()
        #    aeq(batch_, batch)
        #    aeq(k_len_, k_len)
        #    aeq(q_len_ == q_len)
        # END CHECKS

        batch_size = key.size(0)
        dim_per_head = self.dim_per_head
        head_count = self.head_count
        key_len = key.size(1)
        query_len = query.size(1)
        
        attn = None

        def shape(x):
            """Projection."""
            return x.view(batch_size, -1, head_count, dim_per_head) \
                .transpose(1, 2)

        def unshape(x):
            """Compute context."""
            return x.transpose(1, 2).contiguous() \
                    .view(batch_size, -1, head_count * dim_per_head)

        # 1) Project key, value, and query.
        if layer_cache is not None:
            if attn_type == ATTEN_TYPE_SELF:
                query, key, value = self.linear_query(query),\
                                    self.linear_keys(query),\
                                    self.linear_values(query)
                key = shape(key)
                value = shape(value)
                if layer_cache["self_keys"] is not None:
                    key = torch.cat(
                        (layer_cache["self_keys"], key),
                        dim=2)
                if layer_cache["self_values"] is not None:
                    value = torch.cat(
                        (layer_cache["self_values"], value),
                        dim=2)
                layer_cache["self_keys"] = key
                layer_cache["self_values"] = value
            elif attn_type == "context":
                query = self.linear_query(query)
                if layer_cache["memory_keys"] is None:
                    key, value = self.linear_keys(key),\
                                 self.linear_values(value)
                    key = shape(key)
                    value = shape(value)
                else:
                    key, value = layer_cache["memory_keys"],\
                               layer_cache["memory_values"]
                layer_cache["memory_keys"] = key
                layer_cache["memory_values"] = value
        else:
            key = self.linear_keys(key)
            value = self.linear_values(value)
            query = self.linear_query(query)
            key = shape(key)
            value = shape(value)

        if self.max_relative_positions > 0 and attn_type == ATTEN_TYPE_SELF:
            key_len = key.size(2)
            # 1 or key_len x key_len
            relative_positions_matrix = generate_relative_positions_matrix(
                key_len, self.max_relative_positions,
                use_neg_dist=self.use_neg_dist,
                cache=True if layer_cache is not None else False)
            #  1 or key_len x key_len x dim_per_head
            relations_keys = self.relative_positions_embeddings(
                relative_positions_matrix.to(key.device))
            #  1 or key_len x key_len x dim_per_head
            relations_values = self.relative_positions_embeddings(
                relative_positions_matrix.to(key.device))

        query = shape(query)

        key_len = key.size(2)
        query_len = query.size(2)

        if attn_type == ATTEN_TYPE_SELF:
            # 2) Calculate and scale scores.
            query = query / math.sqrt(dim_per_head)
            # batch x num_heads x query_len x key_len
            query_key = torch.matmul(query, key.transpose(2, 3))

            if self.max_relative_positions > 0 and attn_type == ATTEN_TYPE_SELF:
                scores = query_key + relative_matmul(query, relations_keys, True)
            else:
                scores = query_key
            scores = scores.float()

            if mask is not None:
                if len(mask.size()) == 3:
                    mask = mask.unsqueeze(1)  # [B, 1, 1, T_values]
                if len(mask.size()) == 2:
                    mask = mask.unsqueeze(1)
                    mask = mask.unsqueeze(1)  # [B, 1, 1, T_values]
                scores = scores.masked_fill(mask, float("-inf"))
                # scores = scores.masked_fill(mask, -1e18)

            # 3) Apply attention dropout and compute context vectors.
            attn = self.softmax(scores).to(query.dtype)

            if self.training:
                drop_attn = self.dropout(attn)
            else:
                drop_attn = attn

            context_original = torch.matmul(drop_attn, value)
        elif attn_type == ATTEN_TYPE_HYGRA:
            ##################### hydra attention
            ##### TODO: does it need dropout?
            q = query / query.norm(dim=-1, keepdim=True)
            k = key / key.norm(dim=-1, keepdim=True)
                
            ###### origin  
            # ref: https://arxiv.org/pdf/2209.07484.pdf
            # kv = (k * value).sum(dim=-2, keepdim=True)
            # context_original = q * kv
            ###########################
           
            
            qk = torch.matmul(q, k.transpose(2, 3)) 
        
            if self.max_relative_positions > 0 and attn_type == ATTEN_TYPE_SELF:
                scores = qk + relative_matmul(q, relations_keys, True)
            else:
                scores = qk
            scores = scores.float()
        
            if mask is not None:
                if len(mask.size()) == 3:
                    mask = mask.unsqueeze(1)  # [B, 1, 1, T_values]
                if len(mask.size()) == 2:
                    mask = mask.unsqueeze(1)
                    mask = mask.unsqueeze(1)  # [B, 1, 1, T_values]
                scores = scores.masked_fill(mask, float("-inf"))
                # scores = scores.masked_fill(mask, -1e18)

            # 3) Apply attention dropout and compute context vectors.
            attn = self.softmax(scores).to(query.dtype)

            if self.training:
                drop_attn = self.dropout(attn)
            else:
                drop_attn = attn

            context_original = torch.matmul(drop_attn, value)
            
        else:
            assert "Atten_type[{}] is not supported".format(attn_type)

        if self.max_relative_positions > 0 and attn_type == ATTEN_TYPE_SELF:
            context = unshape(context_original
                              + relative_matmul(drop_attn,
                                                relations_values,
                                                False))
        else:
            context = unshape(context_original)

        output = self.final_linear(context)
        # CHECK
        # batch_, q_len_, d_ = output.size()
        # aeq(q_len, q_len_)
        # aeq(batch, batch_)
        # aeq(d, d_)

        # Return multi-head attn
        if attn is not None:
            attns = attn \
                .view(batch_size, head_count,
                    query_len, key_len)
            
            # attn_per_head = [attn.squeeze(1)
                        #  for attn in attn.chunk(head_count, dim=1)]
            
            # print('---', len(attn_per_head),(attn_per_head[0]).shape, attns.shape)
        else:
            attns = None

        return output, attns

    def update_dropout(self, dropout):
        self.dropout.p = dropout

        
if __name__ == "__main__":
    from aicmder.torch import init_seeds
    from misc import sequence_mask
    init_seeds(0)
    heads = 2
    d_model = 4
    dropout = 0
    max_relative_positions = 32
    use_neg_dist = True
    attention = MultiHeadedAttention(head_count=heads, model_dim=d_model,
                                            dropout=dropout,
                                            max_relative_positions=max_relative_positions,
                                            use_neg_dist=use_neg_dist)

    # inputs = torch.randint(0, 20000, (32, 5, d_model))
    N = 3
    inputs = torch.rand(N, 2, d_model, dtype=torch.float32)
    code_len = torch.zeros(N, dtype=torch.long)
    code_len[:] = 2
    print(inputs.shape)
    mask = None if code_len is None else \
            ~sequence_mask(code_len, inputs.shape[1]).unsqueeze(1)
    # context, attn_per_head = attention(inputs, inputs, inputs, mask=mask, attn_type=ATTEN_TYPE_SELF)
    context, attn_per_head = attention(inputs, inputs, inputs, mask=mask, attn_type=ATTEN_TYPE_HYGRA)
    print(type(context), type(attn_per_head))
    print(context.shape, attn_per_head)
    print(torch.nn.MultiheadAttention)
    # import torch.nn as nn
    # import torch

    embed_dim = 3
    num_heads = 1

    x = [
        [1, 0, 1], # Input 1
        [0, 2, 0], # Input 2
        [1, 1, 1] # Input 3
    ]

    x = torch.tensor(x, dtype=torch.float32)

    w_key = [
        [0, 0, 1],
        [1, 1, 0],
        # [0, 1, 0],
        [1, 1, 0]
    ]

    w_query = [
        [1, 0, 1],
        [1, 0, 0],
        # [0, 0, 1],
        [0, 1, 1]
    ]

    w_value = [
        [0 ,0 ,0],
        [1 ,0 ,1],
        # [1 ,0 ,0],
        [0 ,1 ,0]
    ]

    w_key   = torch.tensor(w_key   , dtype=torch.float32)
    w_query = torch.tensor(w_query , dtype=torch.float32)
    w_value = torch.tensor(w_value , dtype=torch.float32)

    # The input to the MultiheadAttention layer is of shape (S,N,E) where S is the sequence length,
    # N is the batch size and E is the embedding dimension. In this example we have only one input vector,
    # so we need to add an extra dimension for the batch size.

    x = x.unsqueeze(1) # shape (3 ,4) -> (3 ,1 ,4)

    # The weights for the projection matrices are stored as parameters of the MultiheadAttention layer,
    # so we need to create an instance of this layer and assign our weights to it.

    multihead_attn = nn.MultiheadAttention(embed_dim=embed_dim,num_heads=num_heads)

    project_weight = torch.cat([w_query,w_key,w_value],dim=0)
    print(project_weight.shape, w_query.shape, w_key.shape, w_value.shape, multihead_attn.in_proj_weight.data.shape)
    multihead_attn.in_proj_weight.data.copy_(project_weight) # shape (12 ,4)

    # We can now call the forward method of the layer with our input vector as query,key and value.

    print(x.shape)
    attn_output,_=multihead_attn(x,x,x)

    print(attn_output.squeeze(1))


import torch
import torch.nn as nn


class RotaryEmbedding(nn.Module):
    def __init__(self, dim, original_impl=False, device=None, dtype=None):
        super().__init__()
        inv_freq = 1.0 / (10000 ** (torch.arange(0, dim, 2, device=device).to(dtype=dtype) / dim))
        self.register_buffer("inv_freq", inv_freq)
        self.dim = dim
        self.original_impl = original_impl

    def forward_impl(
            self, seq_len: int, n_elem: int, dtype: torch.dtype, device: torch.device, base: int = 10000
    ):
        """Enhanced Transformer with Rotary Position Embedding.

        Derived from: https://github.com/labmlai/annotated_deep_learning_paper_implementations/blob/master/labml_nn/
        transformers/rope/__init__.py. MIT License:
        https://github.com/labmlai/annotated_deep_learning_paper_implementations/blob/master/license.
        """
        # $\Theta = {\theta_i = 10000^{\frac{2(i-1)}{d}}, i \in [1, 2, ..., \frac{d}{2}]}$
        theta = 1.0 / (base ** (torch.arange(0, n_elem, 2, dtype=dtype, device=device) / n_elem))

        # Create position indexes `[0, 1, ..., seq_len - 1]`
        seq_idx = torch.arange(seq_len, dtype=dtype, device=device)

        # Calculate the product of position index and $\theta_i$
        idx_theta = torch.outer(seq_idx, theta).float()

        cache = torch.stack([torch.cos(idx_theta), torch.sin(idx_theta)], dim=-1)

        # this is to mimic the behaviour of complex32, else we will get different results
        if dtype in (torch.float16, torch.bfloat16, torch.int8):
            cache = cache.bfloat16() if dtype == torch.bfloat16 else cache.half()
        return cache

    def forward(self, max_seq_len, offset=0):
        return self.forward_impl(
            max_seq_len, self.dim, dtype=self.inv_freq.dtype, device=self.inv_freq.device
        )


@torch.jit.script
def apply_rotary_pos_emb(x: torch.Tensor, rope_cache: torch.Tensor) -> torch.Tensor:
    # x: [sq, b, np, hn]
    sq, b, np, hn = x.size(0), x.size(1), x.size(2), x.size(3)
    rot_dim = rope_cache.shape[-2] * 2
    x, x_pass = x[..., :rot_dim], x[..., rot_dim:]
    # truncate to support variable sizes
    rope_cache = rope_cache[:sq]
    xshaped = x.reshape(sq, -1, np, rot_dim // 2, 2)
    rope_cache = rope_cache.view(sq, -1, 1, xshaped.size(3), 2)
    x_out2 = torch.stack(
        [
            xshaped[..., 0] * rope_cache[..., 0] - xshaped[..., 1] * rope_cache[..., 1],
            xshaped[..., 1] * rope_cache[..., 0] + xshaped[..., 0] * rope_cache[..., 1],
        ],
        -1,
    )
    x_out2 = x_out2.flatten(3)
    return torch.cat((x_out2, x_pass), dim=-1)


if __name__ == "__main__":
    config = {
        "_name_or_path": "THUDM/chatglm2-6b",
        "model_type": "chatglm",
        "architectures": [
            "ChatGLMModel"
        ],
        "auto_map": {
            "AutoConfig": "configuration_chatglm.ChatGLMConfig",
            "AutoModel": "modeling_chatglm.ChatGLMForConditionalGeneration",
            "AutoModelForSeq2SeqLM": "modeling_chatglm.ChatGLMForConditionalGeneration"
        },
        "add_bias_linear": False,
        "add_qkv_bias": True,
        "apply_query_key_layer_scaling": True,
        "apply_residual_connection_post_layernorm": False,
        "attention_dropout": 0.0,
        "attention_softmax_in_fp32": True,
        "bias_dropout_fusion": True,
        "ffn_hidden_size": 13696,
        "fp32_residual_connection": False,
        "hidden_dropout": 0.0,
        "hidden_size": 4096,
        "kv_channels": 128,
        "layernorm_epsilon": 1e-05,
        "multi_query_attention": True,
        "multi_query_group_num": 2,
        "num_attention_heads": 32,
        "num_layers": 28,
        "original_rope": True,
        "padded_vocab_size": 65024,
        "post_layer_norm": True,
        "rmsnorm": True,
        "seq_length": 32768,
        "use_cache": True,
        "torch_dtype": "float16",
        "transformers_version": "4.27.1",
        "tie_word_embeddings": False,
        "eos_token_id": 2,
        "pad_token_id": 0
    }
    from aicmder.utils import AttrDict
    config = AttrDict(config)
    rotary_dim = (config.hidden_size // config.num_attention_heads
                  if config.kv_channels is None else config.kv_channels)
    
    torch_dtype = getattr(torch, config.torch_dtype)
    print(rotary_dim, torch_dtype)
    device = torch.device("cuda:0")
    rotary_pos_emb = RotaryEmbedding(
        rotary_dim // 2, original_impl=config.original_rope, device=device, dtype=torch_dtype)

    seq_length = 10
    cache = rotary_pos_emb(seq_length)
    print(cache, cache.shape)



# https://github.dev/OpenNMT/OpenNMT-py/blob/173e113b1220b32c1d03a0f8d3f9d44759c071de/onmt/decoders/transformer.py#L12
import torch
import torch.nn as nn
from aicmder.torch.transformer.util_class import ActivationFunction
from aicmder.torch.transformer import AverageAttention, \
PositionwiseFeedForward, sequence_mask
try:
    from multi_headed_attn import MultiHeadedAttention
except:
    from aicmder.torch.transformer import MultiHeadedAttention
from aicmder.torch.transformer.embedding import PADDING_INDEX
from aicmder.torch.transformer.multi_headed_attn import ATTEN_TYPE_SELF, ATTEN_TYPE_HYGRA
from aicmder.torch.transformer.base_model import DecoderBase

class DecoderLayerBase(nn.Module):

    def __init__(self,
                 d_model,
                 heads,
                 d_ff,
                 dropout,
                 attention_dropout,
                 self_attn_type="scaled-dot",
                 max_relative_positions=0,
                 aan_useffn=False,
                 full_context_alignment=False,
                 alignment_heads=0,
                 pos_ffn_activation_fn=ActivationFunction.relu):
        """
        Args:
            d_model (int): the dimension of keys/values/queries in
                :class:`MultiHeadedAttention`, also the input size of
                the first-layer of the :class:`PositionwiseFeedForward`.
            heads (int): the number of heads for MultiHeadedAttention.
            d_ff (int): the second-layer of the
                :class:`PositionwiseFeedForward`.
            dropout (float): dropout in residual, self-attn(dot) and
                feed-forward
            attention_dropout (float): dropout in context_attn  (and
                self-attn(avg))
            self_attn_type (string): type of self-attention scaled-dot,
                average
            max_relative_positions (int):
                Max distance between inputs in relative positions
                representations
            aan_useffn (bool): Turn on the FFN layer in the AAN decoder
            full_context_alignment (bool):
                whether enable an extra full context decoder forward for
                alignment
            alignment_heads (int):
                N. of cross attention heads to use for alignment guiding
            pos_ffn_activation_fn (ActivationFunction):
                activation function choice for PositionwiseFeedForward layer

        """
        super(DecoderLayerBase, self).__init__()

        if self_attn_type == "scaled-dot":
            self.self_attn = MultiHeadedAttention(
                head_count=heads,
                model_dim=d_model,
                dropout=attention_dropout,
                max_relative_positions=max_relative_positions,
            )
        elif self_attn_type == "average":
            self.self_attn = AverageAttention(
                d_model, dropout=attention_dropout, aan_useffn=aan_useffn
            )

        self.feed_forward = PositionwiseFeedForward(d_model, d_ff, dropout,
                                                    pos_ffn_activation_fn
                                                    )

        self.layer_norm_1 = nn.LayerNorm(d_model, eps=1e-6)
        # self.layer_norm = LayerNorm(d_model)

        self.drop = nn.Dropout(dropout)
        self.full_context_alignment = full_context_alignment
        self.alignment_heads = alignment_heads

    def forward(self, *args, **kwargs):
        """Extend `_forward` for (possibly) multiple decoder pass:
        Always a default (future masked) decoder forward pass,
        Possibly a second future aware decoder pass for joint learn
        full context alignement, :cite:`garg2019jointly`.

        Args:
            * All arguments of _forward.
            with_align (bool): whether return alignment attention.

        Returns:
            (FloatTensor, FloatTensor, FloatTensor or None):

            * output ``(batch_size, T, model_dim)``
            * top_attn ``(batch_size, T, src_len)``
            * attn_align ``(batch_size, T, src_len)`` or None
        """
        with_align = kwargs.pop("with_align", False)
        output, attns = self._forward(*args, **kwargs)
        if attns is not None:
            top_attn = attns[:, 0, :, :].contiguous()
        else:
            top_attn = attns
        attn_align = None
        if with_align:
            if self.full_context_alignment:
                # return _, (B, Q_len, K_len)
                _, attns = self._forward(*args, **kwargs, future=True)

            if self.alignment_heads > 0:
                attns = attns[:, : self.alignment_heads, :, :].contiguous()
            # layer average attention across heads, get ``(B, Q, K)``
            # Case 1: no full_context, no align heads -> layer avg baseline
            # Case 2: no full_context, 1 align heads -> guided align
            # Case 3: full_context, 1 align heads -> full cte guided align
            attn_align = attns.mean(dim=1)
        return output, top_attn, attn_align

    def update_dropout(self, dropout, attention_dropout):
        self.self_attn.update_dropout(attention_dropout)
        self.feed_forward.update_dropout(dropout)
        self.drop.p = dropout

    def _forward(self, *args, **kwargs):
        raise NotImplementedError

    def _compute_dec_mask(self, tgt_pad_mask, future):
        tgt_len = tgt_pad_mask.size(-1)
        if not future:  # apply future_mask, result mask in (B, T, T)
            future_mask = torch.ones(
                [tgt_len, tgt_len],
                device=tgt_pad_mask.device,
                dtype=torch.uint8,
            )
            future_mask = future_mask.triu_(1).view(1, tgt_len, tgt_len)
            # BoolTensor was introduced in pytorch 1.2
            try:
                future_mask = future_mask.bool()
            except AttributeError:
                pass
            dec_mask = torch.gt(tgt_pad_mask + future_mask, 0)
        else:  # only mask padding, result mask in (B, 1, T)
            dec_mask = tgt_pad_mask
        return dec_mask

    def _forward_self_attn(self, inputs_norm, dec_mask, layer_cache, step, attn_type=ATTEN_TYPE_SELF):
        if isinstance(self.self_attn, MultiHeadedAttention):
            return self.self_attn(
                inputs_norm,
                inputs_norm,
                inputs_norm,
                mask=dec_mask,
                layer_cache=layer_cache,
                attn_type=attn_type,
            )
        elif isinstance(self.self_attn, AverageAttention):
            return self.self_attn(
                inputs_norm, mask=dec_mask, layer_cache=layer_cache, step=step
            )
        else:
            raise ValueError(
                f"self attention {type(self.self_attn)} not supported"
            )



class DecoderLayer(DecoderLayerBase):
    """Transformer Decoder layer block in Pre-Norm style.
    Pre-Norm style is an improvement w.r.t. Original paper's Post-Norm style,
    providing better converge speed and performance. This is also the actual
    implementation in tensor2tensor and also avalable in fairseq.
    See https://tunz.kr/post/4 and :cite:`DeeperTransformer`.

    ```mermaid
        graph LR
        %% "*SubLayer" can be self-attn, src-attn or feed forward block
            A(input) --> B[Norm]
            B --> C["*SubLayer"]
            C --> D[Drop]
            D --> E((+))
            A --> E
            E --> F(out)
    ```
    
    """
    
    def __init__(
        self,
        d_model,
        heads,
        d_ff,
        dropout,
        attention_dropout,
        self_attn_type="scaled-dot",
        max_relative_positions=0,
        aan_useffn=False,
        full_context_alignment=False,
        alignment_heads=0,
        pos_ffn_activation_fn=ActivationFunction.relu,
    ):
        """
        Args:
            See TransformerDecoderLayerBase
        """
        super(DecoderLayer, self).__init__(
            d_model,
            heads,
            d_ff,
            dropout,
            attention_dropout,
            self_attn_type,
            max_relative_positions,
            aan_useffn,
            full_context_alignment,
            alignment_heads,
            pos_ffn_activation_fn=pos_ffn_activation_fn,
        )
        self.context_attn = MultiHeadedAttention(
            head_count=heads, model_dim=d_model, dropout=attention_dropout,
            max_relative_positions=max_relative_positions
        )
        self.layer_norm_2 = nn.LayerNorm(d_model, eps=1e-6)

    def update_dropout(self, dropout, attention_dropout):
        super(DecoderLayer, self).update_dropout(
            dropout, attention_dropout
        )
        self.context_attn.update_dropout(attention_dropout)
        

    def _forward(
        self,
        inputs,
        memory_bank,
        src_pad_mask,
        tgt_pad_mask,
        attn_type,
        layer_cache=None,
        step=None,
        future=False,
    ):
        """A naive forward pass for transformer decoder.

        # T: could be 1 in the case of stepwise decoding or tgt_len

        Args:
            inputs (FloatTensor): ``(batch_size, T, model_dim)``
            memory_bank (FloatTensor): ``(batch_size, src_len, model_dim)``
            src_pad_mask (bool): ``(batch_size, 1, src_len)``
            tgt_pad_mask (bool): ``(batch_size, 1, T)``
            layer_cache (dict or None): cached layer info when stepwise decode
            step (int or None): stepwise decoding counter
            future (bool): If set True, do not apply future_mask.

        Returns:
            (FloatTensor, FloatTensor):

            * output ``(batch_size, T, model_dim)``
            * attns ``(batch_size, head, T, src_len)``

        """
        dec_mask = None

        if inputs.size(1) > 1:
            # masking is necessary when sequence length is greater than one
            dec_mask = self._compute_dec_mask(tgt_pad_mask, future)

        # apply layer norm first
        inputs_norm = self.layer_norm_1(inputs)

        query, _ = self._forward_self_attn(
            inputs_norm, dec_mask, layer_cache, step, attn_type=attn_type
        )

        query = self.drop(query) + inputs

        query_norm = self.layer_norm_2(query)
        mid, attns = self.context_attn(
            memory_bank,
            memory_bank,
            query_norm,
            mask=src_pad_mask,
            layer_cache=layer_cache,
            attn_type=attn_type,
        )
        output = self.feed_forward(self.drop(mid) + query)

        return output, attns
    
    

class TransformerDecoderBase(DecoderBase):
    def __init__(self, d_model, copy_attn, embeddings, alignment_layer):
        super(TransformerDecoderBase, self).__init__()

        self.embeddings = embeddings

        # Decoder State
        self.state = {}

        # previously, there was a GlobalAttention module here for copy
        # attention. But it was never actually used -- the "copy" attention
        # just reuses the context attention.
        self._copy = copy_attn
        self.layer_norm = nn.LayerNorm(d_model, eps=1e-6)

        self.alignment_layer = alignment_layer

    @classmethod
    def from_opt(cls, opt, embeddings):
        """Alternate constructor."""
        return cls(
            opt.dec_layers,
            opt.dec_rnn_size,
            opt.heads,
            opt.transformer_ff,
            opt.copy_attn,
            opt.self_attn_type,
            opt.dropout[0] if type(opt.dropout) is list else opt.dropout,
            opt.attention_dropout[0]
            if type(opt.attention_dropout) is list
            else opt.attention_dropout,
            embeddings,
            opt.max_relative_positions,
            opt.aan_useffn,
            opt.full_context_alignment,
            opt.alignment_layer,
            alignment_heads=opt.alignment_heads,
            pos_ffn_activation_fn=opt.pos_ffn_activation_fn,
        )

    def init_state(self, src_len, max_len):
        """Initialize decoder state."""
        self.state["src_len"] = src_len # [B]
        self.state["src_max_len"] = max_len  # an integer
        self.state["cache"] = None

    def map_state(self, fn):
        def _recursive_map(struct, batch_dim=0):
            for k, v in struct.items():
                if v is not None:
                    if isinstance(v, dict):
                        _recursive_map(v)
                    else:
                        struct[k] = fn(v, batch_dim)

        if self.state["src"] is not None:
            self.state["src"] = fn(self.state["src"], 1)
        if self.state["cache"] is not None:
            _recursive_map(self.state["cache"])


    def forward(self, *args, **kwargs):
        raise NotImplementedError

    def update_dropout(self, dropout, attention_dropout):
        self.embeddings.update_dropout(dropout)
        for layer in self.transformer_layers:
            layer.update_dropout(dropout, attention_dropout)
            


class TransformerDecoder(TransformerDecoderBase):
    """The Transformer decoder from "Attention is All You Need".
    :cite:`DBLP:journals/corr/VaswaniSPUJGKP17`

    .. mermaid::

       graph BT
          A[input]
          B[multi-head self-attn]
          BB[multi-head src-attn]
          C[feed forward]
          O[output]
          A --> B
          B --> BB
          BB --> C
          C --> O


    Args:
        num_layers (int): number of decoder layers.
        d_model (int): size of the model
        heads (int): number of heads
        d_ff (int): size of the inner FF layer
        copy_attn (bool): if using a separate copy attention
        self_attn_type (str): type of self-attention scaled-dot, average
        dropout (float): dropout in residual, self-attn(dot) and feed-forward
        attention_dropout (float): dropout in context_attn (and self-attn(avg))
        embeddings (onmt.modules.Embeddings):
            embeddings to use, should have positional encodings
        max_relative_positions (int):
            Max distance between inputs in relative positions representations
        aan_useffn (bool): Turn on the FFN layer in the AAN decoder
        full_context_alignment (bool):
            whether enable an extra full context decoder forward for alignment
        alignment_layer (int): N° Layer to supervise with for alignment guiding
        alignment_heads (int):
            N. of cross attention heads to use for alignment guiding
    """

    def __init__(
        self,
        num_layers,
        d_model,
        heads,
        d_ff,
        embeddings=None,
        max_relative_positions=0,
        copy_attn=False,
        dropout=0.2,
        attention_dropout=0.2,
        aan_useffn=False,
        alignment_layer=None,
        full_context_alignment=False,
        alignment_heads=0,
        self_attn_type="scaled-dot",
        pos_ffn_activation_fn=ActivationFunction.relu,
    ):
        super(TransformerDecoder, self).__init__(
            d_model, copy_attn, embeddings, alignment_layer
        )

        self.transformer_layers = nn.ModuleList(
            [
                DecoderLayer(
                    d_model,
                    heads,
                    d_ff,
                    dropout,
                    attention_dropout,
                    self_attn_type=self_attn_type,
                    max_relative_positions=max_relative_positions,
                    aan_useffn=aan_useffn,
                    full_context_alignment=full_context_alignment,
                    alignment_heads=alignment_heads,
                    pos_ffn_activation_fn=pos_ffn_activation_fn,
                )
                for i in range(num_layers)
            ]
        )


    def forward(self, tgt, tgt_token_type_ids=None, memory_bank=None, src_pad_mask=None,  tgt_pad_mask=None, step=None, attn_type=ATTEN_TYPE_SELF, **kwargs):
        """Decode, possibly stepwise."""
        if memory_bank is None:
            memory_bank = self.embeddings(tgt)
        if step == 0:
            self._init_cache(memory_bank)

        # tgt_words = tgt[:, :, 0].transpose(0, 1)
        # tgt_words = tgt[:, :, 0]

        if self.embeddings is not None:
            if tgt_token_type_ids is None:
                tgt_token_type_ids = torch.ones(tgt.shape, dtype=torch.long)
            if tgt.dim() == 2:
                tgt = tgt.unsqueeze(2)
            emb = self.embeddings(tgt, step=0, token_type_ids=tgt_token_type_ids)

            # emb = self.embeddings(tgt, step=step)
            pad_idx = self.embeddings.word_padding_idx if self.embeddings is not None else PADDING_INDEX
            tgt_pad_mask = tgt.data.eq(pad_idx)
        else:
            emb = tgt
            assert tgt_pad_mask is not None, "target padding should be given."
        assert emb.dim() == 3  # len x batch x embedding_dim

        output = emb
        # output = emb.transpose(0, 1).contiguous()
        # src_memory_bank = memory_bank.transpose(0, 1).contiguous()


        # src_lens = kwargs["memory_lengths"]
        # src_max_len = self.state["src"].shape[0]

        if src_pad_mask is None:
            src_lens = self.state["src_len"]
            src_max_len = self.state["src_max_len"] 
            src_pad_mask = ~sequence_mask(src_lens, src_max_len).unsqueeze(1)
        
        if len(tgt_pad_mask.shape) == 2:
            tgt_pad_mask = tgt_pad_mask.unsqueeze(1)  # [B, 1, T_tgt]
            
        with_align = kwargs.pop("with_align", False)
        attn_aligns = []

        for i, layer in enumerate(self.transformer_layers):
            layer_cache = (
                self.state["cache"]["layer_{}".format(i)]
                if step is not None
                else None
            )
            src_memory_bank = memory_bank[i] if isinstance(memory_bank, list) else memory_bank
            output, attn, attn_align = layer(
                output,
                src_memory_bank,
                src_pad_mask,
                tgt_pad_mask,
                layer_cache=layer_cache,
                step=step,
                with_align=with_align,
                attn_type=attn_type
            )
            if attn_align is not None:
                attn_aligns.append(attn_align)

        output = self.layer_norm(output)
        # dec_outs = output.transpose(0, 1).contiguous()
        # attn = attn.transpose(0, 1).contiguous()
        dec_outs = output.contiguous()
        if attn is not None:
            attn = attn.contiguous()

        attns = {"std": attn}
        if self._copy:
            attns["copy"] = attn
        if with_align and self.alignment_layer is not None:
            attns["align"] = attn_aligns[self.alignment_layer]  # `(B, Q, K)`
            # attns["align"] = torch.stack(attn_aligns, 0).mean(0)  # All avg

        # TODO change the way attns is returned dict => list or tuple (onnx)
        return dec_outs, attns

    def _init_cache(self, memory_bank):
        self.state["cache"] = {}
        batch_size = memory_bank.size(1)
        depth = memory_bank.size(-1)

        for i, layer in enumerate(self.transformer_layers):
            layer_cache = {"memory_keys": None, "memory_values": None}
            if isinstance(layer.self_attn, AverageAttention):
                layer_cache["prev_g"] = torch.zeros(
                    (batch_size, 1, depth), device=memory_bank.device
                )
            else:
                layer_cache["self_keys"] = None
                layer_cache["self_values"] = None
            self.state["cache"]["layer_{}".format(i)] = layer_cache


if __name__ == "__main__":
    from aicmder.torch import init_seeds
    from aicmder.torch.transformer import sequence_mask, Embeddings
    import torch
    init_seeds(0)
    nlayers = 6
    input_size = 512
    trans_drop = 0.2
    num_head = 8
    d_ff = 2048
    coverage_attn = False

    tgt_word_embeddings = Embeddings(input_size, 184866, 0, position_encoding="default")
    decoder = TransformerDecoder(
                num_layers=nlayers,
                d_model=input_size,
                heads=num_head,
                d_ff=d_ff,
                copy_attn=False,
                embeddings=tgt_word_embeddings,
                # coverage_attn=coverage_attn,
                dropout=trans_drop
            ) 

    memory_bank = torch.rand(32, 5, input_size, dtype=torch.float32)
    # tgt = torch.rand(32, 7, dtype=torch.int64)
    tgt = torch.randint(0, 2000, (32, 7))

    max_len = memory_bank[0].shape[1] \
            if isinstance(memory_bank, list) else memory_bank.shape[1]
    src_len = [5] * 32
    src_len = torch.tensor(src_len)
    print(src_len.shape)
    decoder.init_state(src_len, max_len)
    dec_outs, attns = decoder(tgt, memory_bank=memory_bank, attn_type=ATTEN_TYPE_HYGRA)
    print(dec_outs.shape)


import logging
import os
import torch.nn as nn
import torch

VERBOSE = str(os.getenv('TRAIN_VERBOSE', True)
              ).lower() == 'true'  # global verbose mode


def set_logging(name=None, verbose=VERBOSE):
    # Sets level and returns logger
    # if is_kaggle():
    #     for h in logging.root.handlers:
    #         logging.root.removeHandler(h)  # remove all handlers associated with the root logger object
    rank = int(os.getenv('RANK', -1))  # rank in world for Multi-GPU trainings
    level = logging.INFO if verbose and rank in {-1, 0} else logging.ERROR
    log = logging.getLogger(name)
    log.setLevel(level)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(message)s"))
    handler.setLevel(level)
    log.addHandler(handler)


def colorstr(*input):
    # Colors a string https://en.wikipedia.org/wiki/ANSI_escape_code, i.e.  colorstr('blue', 'hello world')
    # color arguments, string
    *args, string = input if len(input) > 1 else ('blue', 'bold', input[0])
    colors = {
        'black': '\033[30m',  # basic colors
        'red': '\033[31m',
        'green': '\033[32m',
        'yellow': '\033[33m',
        'blue': '\033[34m',
        'magenta': '\033[35m',
        'cyan': '\033[36m',
        'white': '\033[37m',
        'bright_black': '\033[90m',  # bright colors
        'bright_red': '\033[91m',
        'bright_green': '\033[92m',
        'bright_yellow': '\033[93m',
        'bright_blue': '\033[94m',
        'bright_magenta': '\033[95m',
        'bright_cyan': '\033[96m',
        'bright_white': '\033[97m',
        'end': '\033[0m',  # misc
        'bold': '\033[1m',
        'underline': '\033[4m'}
    return ''.join(colors[x] for x in args) + f'{string}' + colors['end']


set_logging()  # run before defining LOGGER
# define globally (used in train.py, val.py, detect.py, etc.)
LOGGER = logging.getLogger("train")




# src: https://github.com/OpenNMT/OpenNMT-py/blob/master/onmt/decoders/decoder.py
from aicmder.torch.transformer.base_model import DecoderBase
import torch.nn as nn
from aicmder.torch.transformer.misc import aeq
from aicmder.torch.transformer.state import RNNDecoderState
import torch
from aicmder.torch.transformer.global_attention import GlobalAttention

def rnn_factory(rnn_type, **kwargs):
    """ rnn factory, Use pytorch version when available. """
    no_pack_padded_seq = False
    if rnn_type == "SRU":
        # SRU doesn't support PackedSequence.
        no_pack_padded_seq = True
        # rnn = onmt.models.sru.SRU(**kwargs)
        raise NotImplementedError("Not implemented yet!")
    else:
        rnn = getattr(nn, rnn_type)(**kwargs)
    return rnn, no_pack_padded_seq


# many part of the codes are copied from OpenNMT-Py sources
class RNNDecoderBase(nn.Module):
    """
    Base recurrent attention-based decoder class.

    .. mermaid::
       graph BT
          A[Input]
          subgraph RNN
             C[Pos 1]
             D[Pos 2]
             E[Pos N]
          end
          G[Decoder State]
          H[Decoder State]
          I[Outputs]
          F[Memory_Bank]
          A--emb-->C
          A--emb-->D
          A--emb-->E
          H-->C
          C-- attn --- F
          D-- attn --- F
          E-- attn --- F
          C-->I
          D-->I
          E-->I
          E-->G
          F---I

    Args:
       rnn_type (:obj:`str`):
          style of recurrent unit to use, one of [LSTM, GRU]
       bidirectional (bool) : use with a bidirectional encoder
       num_layers (int) : number of stacked layers
       hidden_size (int) : hidden size of each layer
       attn_type (str) : see :obj:`nqa.modules.GlobalAttention`
       dropout (float) : dropout value for :obj:`nn.Dropout`
    """

    def __init__(self,
                 input_size,
                 bidirectional_encoder,
                 num_layers,
                 hidden_size,
                 rnn_type="LSTM",
                 attn_type='general',
                 coverage_attn=False,
                 copy_attn=False,
                 reuse_copy_attn=False,
                 dropout=0.0):

        super(RNNDecoderBase, self).__init__()

        # Basic attributes.
        self.decoder_type = 'rnn'
        self.bidirectional_encoder = bidirectional_encoder
        self.num_layers = num_layers
        self.hidden_size = hidden_size
        self.dropout = nn.Dropout(dropout)
        self.input_size = input_size
        self.attn_type = attn_type

        # Build the RNN.
        kwargs = {'input_size': input_size,
                  'hidden_size': hidden_size,
                  'num_layers': num_layers,
                  'dropout': dropout,
                  'batch_first': True}
        self.rnn = getattr(nn, rnn_type)(**kwargs)

        # Set up the standard attention.
        self._coverage = coverage_attn
        self.attn = None
        if self.attn_type:
            self.attn = GlobalAttention(
                hidden_size, coverage=coverage_attn,
                attn_type=self.attn_type
            )
        else:
            assert not self._coverage
            if copy_attn and reuse_copy_attn:
                raise RuntimeError('Attn is turned off, so reuse_copy_attn flag must be false')

        # Set up a separated copy attention layer, if needed.
        self._copy = copy_attn
        self._reuse_copy_attn = reuse_copy_attn
        self.copy_attn = None
        if copy_attn and not reuse_copy_attn:
            self.copy_attn = GlobalAttention(
                hidden_size, attn_type=self.attn_type
            )

    def count_parameters(self):
        params = list(self.rnn.parameters())
        if self.attn is not None:
            params = params + list(self.attn.parameters())
        if self.copy_attn is not None:
            params = params + list(self.copy_attn.parameters())
        return sum(p.numel() for p in params if p.requires_grad)

    def forward(self, tgt, memory_bank, state=None, memory_lengths=None, encoder_final=None):
        """
        Args:
            tgt (`LongTensor`): sequences of padded tokens
                 `[batch x tgt_len x nfeats]`.
            memory_bank (`FloatTensor`): vectors from the encoder
                 `[batch x src_len x hidden]`.
            state (:obj:`onmt.models.DecoderState`):
                 decoder state object to initialize the decoder
            memory_lengths (`LongTensor`): the padded source lengths
                `[batch]`.
        Returns:
            (`FloatTensor`,:obj:`onmt.Models.DecoderState`,`FloatTensor`):
                * decoder_outputs: output from the decoder (after attn)
                         `[batch x tgt_len x hidden]`.
                * decoder_state: final hidden state from the decoder
                * attns: distribution over src at each tgt
                        `[batch x tgt_len x src_len]`.
        """
        assert state is not None or encoder_final is not None
        if encoder_final is not None and state is None:
            state = self.init_decoder_state(encoder_final)
        # Check
        assert isinstance(state, RNNDecoderState)
        # tgt.size() returns tgt length and batch
        tgt_batch, _, _ = tgt.size()
        if self.attn is not None:
            memory_batch, _, _ = memory_bank.size()
            aeq(tgt_batch, memory_batch)
        # END

        # Run the forward pass of the RNN.
        decoder_final, decoder_outputs, attns = self._run_forward_pass(
            tgt, memory_bank, state, memory_lengths=memory_lengths)

        coverage = None
        if "coverage" in attns:
            coverage = attns["coverage"]
        # Update the state with the result.
        state.update_state(decoder_final, coverage)

        return decoder_outputs, state, attns

    def init_decoder_state(self, encoder_final):
        """ Init decoder state with last state of the encoder """

        def _fix_enc_hidden(hidden):
            # The encoder hidden is  (layers*directions) x batch x dim.
            # We need to convert it to layers x batch x (directions*dim).
            if self.bidirectional_encoder:
                hidden = torch.cat([hidden[0:hidden.size(0):2],
                                    hidden[1:hidden.size(0):2]], 2)
            return hidden

        if isinstance(encoder_final, tuple):  # LSTM
            return RNNDecoderState(self.hidden_size,
                                   tuple([_fix_enc_hidden(enc_hid)
                                          for enc_hid in encoder_final]))
        else:  # GRU
            return RNNDecoderState(self.hidden_size,
                                   _fix_enc_hidden(encoder_final))


class RNNDecoder(RNNDecoderBase):
    """
    Standard fully batched RNN decoder with attention.
    Faster implementation, uses CuDNN for implementation.
    See :obj:`RNNDecoderBase` for options.
    Based around the approach from
    "Neural Machine Translation By Jointly Learning To Align and Translate"
    :cite:`Bahdanau2015`
    """

    def _run_forward_pass(self, tgt, memory_bank, state, memory_lengths=None):
        """
        Private helper for running the specific RNN forward pass.
        Must be overriden by all subclasses.
        Args:
            tgt (LongTensor): a sequence of input tokens tensors
                                 [batch x len x nfeats].
            memory_bank (FloatTensor): output(tensor sequence) from the encoder
                        RNN of size (batch x src_len x hidden_size).
            state (FloatTensor): hidden state from the encoder RNN for
                                 initializing the decoder.
            memory_lengths (LongTensor): the source memory_bank lengths.
        Returns:
            decoder_final (Tensor): final hidden state from the decoder.
            decoder_outputs (Tensor): output from the decoder (after attn)
                         `[batch x tgt_len x hidden]`.
            attns (Tensor): distribution over src at each tgt
                        `[batch x tgt_len x src_len]`.
        """
        # Initialize local and return variables.
        attns = {}

        emb = tgt
        assert emb.dim() == 3

        coverage = state.coverage

        if isinstance(self.rnn, nn.GRU):
            rnn_output, decoder_final = self.rnn(emb, state.hidden[0])
        else:
            rnn_output, decoder_final = self.rnn(emb, state.hidden)

        # Check
        tgt_batch, tgt_len, _ = tgt.size()
        output_batch, output_len, _ = rnn_output.size()
        aeq(tgt_len, output_len)
        aeq(tgt_batch, output_batch)
        # END

        # Calculate the attention.
        if self.attn is not None:
            decoder_outputs, p_attn, coverage_v = self.attn(
                rnn_output.contiguous(),
                memory_bank,
                memory_lengths=memory_lengths,
                coverage=coverage,
                softmax_weights=False
            )
            attns["std"] = p_attn
        else:
            decoder_outputs = rnn_output.contiguous()

        # Update the coverage attention.
        if self._coverage:
            if coverage_v is None:
                coverage = coverage + p_attn \
                    if coverage is not None else p_attn
            else:
                coverage = coverage + coverage_v \
                    if coverage is not None else coverage_v
            attns["coverage"] = coverage

        decoder_outputs = self.dropout(decoder_outputs)
        # Run the forward pass of the copy attention layer.
        if self._copy and not self._reuse_copy_attn:
            _, copy_attn, _ = self.copy_attn(decoder_outputs,
                                             memory_bank,
                                             memory_lengths=memory_lengths,
                                             softmax_weights=False)
            attns["copy"] = copy_attn
        elif self._copy:
            attns["copy"] = attns["std"]

        return decoder_final, decoder_outputs, attns


if __name__ == "__main__":
    from aicmder.torch import init_seeds
    from aicmder.benchmark import model_info
    init_seeds(0)

    rnn_type = "LSTM"
    bidirection = True
    num_layers = 6
    hidden_size = 200
    input_size = 300
    rnn = RNNDecoder(rnn_type=rnn_type, bidirectional_encoder=bidirection, num_layers=num_layers,
                    #  attn_type=None,
                     hidden_size=hidden_size, coverage_attn=False, reuse_copy_attn=False, dropout=0.2, input_size=input_size)

    print(rnn)

    batch_size = 32
    max_len = 7
    # input
    tgt = torch.rand(batch_size, max_len, input_size, dtype=torch.float32)

    # memory
    memory_bank = torch.rand(batch_size, 5, hidden_size, dtype=torch.float32)
    memory_lengths = torch.Tensor([5] * batch_size)

    hidden = (torch.rand(num_layers * 2, batch_size, 100), torch.rand(num_layers * 2, batch_size, 100))
    dec_state = rnn.init_decoder_state(hidden)

    decoder_outputs, state, attns = rnn(tgt,
                                    memory_bank,
                                    # state=dec_state,
                                    encoder_final=hidden,
                                    memory_lengths=memory_lengths)

    print(decoder_outputs.shape, attns['std'].shape if 'std' in attns else None)
    # model_info(rnn, verbose=True)



import os, json
os.environ["CMD_CLIENT_PORT"] = "6655"
os.environ["CMD_WORKER_PORT"] = "6656"
import aicmder as cmder
from aicmder.module.module import serving, moduleinfo
from image_utils import readb64
from face_detection import ImageEmbeddings, SFaceModel, load_embeds
import threading
import time

@moduleinfo(name='face')
class FaceModule(cmder.Module):

    def __init__(self, **kwargs) -> None:
        print("init FaceModule", kwargs)
        # self.model = SFaceModel("/old_faith/home/faith/.deepface/weights/face_recognition_sface_2021dec.onnx")
        self.model = SFaceModel("/home/faith/torch_pt/face_recognition_sface_2021dec.onnx")
        e = load_embeds()
        self.embed = ImageEmbeddings() if e is None else e
        self.face_dir = os.path.abspath(os.path.dirname(__file__))
        if not os.path.exists( self.face_dir):
            os.makedirs(self.face_dir)
        print("current face dir", self.face_dir)
        self.save_thread = threading.Thread(target=self.save_face)
        self.save_thread.start()

    def save_face(self):
        while True:
            self.embed.save()
            # print("saving")
            time.sleep(8)

    @serving
    def predict(self, **kwargs):
        if "method" in kwargs:     
            resp_d = {}
            # print(kwargs["method"], kwargs["method"] == "VerifyFace")
            if "AddFace" == kwargs["method"]:
                img_base64 = kwargs["img"]
                id = kwargs["id"]
                file_path = "{}/face/{}.png".format(self.face_dir , id)
  
                img_bgr = readb64(img_base64, filename=file_path, save=True)
                ret = self.embed.add_file(file_path, self.model)
                resp_d["code"] = 1 if ret else -1
                json_ret = json.dumps(resp_d)
                return json_ret
            if "DelFace" == kwargs["method"]:
                id = kwargs["id"]
                ret = self.embed.remove(id)
                resp_d["code"] = 1 if ret else -1
                json_ret = json.dumps(resp_d)
                return json_ret
            if "VerifyFace" == kwargs["method"]:
                img_base64 = kwargs["img"]
                ret = self.embed.verify_face(img_base64, self.model, base64_img=True)
                if ret is None:
                    ret = resp_d
                json_ret = json.dumps(ret)
                return json_ret
        return '{"data": "-1"}'


# https://github.com/serengil/deepface/issues/351
# from deepface.detectors import FaceDetector
# import cv2

# img_path = "couple.jpg"
# detector_name = "opencv"

# img = cv2.imread(img_path)

# detector = FaceDetector.build_model(detector_name) #set opencv, ssd, dlib, mtcnn or retinaface

# obj = FaceDetector.detect_faces(detector, detector_name, img)

# print("there are ",len(obj)," faces")


# https://github.com/serengil/deepface/issues/419
# from retinaface import RetinaFace
# from deepface import DeepFace

# faces = RetinaFace.extract_faces("img.jpg")
# for face in faces:
#     obj = DeepFace.analyze(img)
#     print(obj["age"])


if __name__ == "__main__":
    with open("face.json") as json_f:
        config = json.load(json_f)
        # print(config)
        exec_cmd = ['-w', config["w"], '-c', json.dumps(config["config"]),
                   '-p', config["port"], '--max_connect', config["max_conn"], '--device_map']
        exec_cmd.extend(config["device"])
        print(exec_cmd)
        serve = cmder.serve.ServeCommand()
        serve.execute(exec_cmd)


# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
PyTorch utils
"""

import datetime
import math
import os
import platform
import subprocess
import time
from contextlib import contextmanager
from copy import deepcopy
from pathlib import Path

import torch
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F

# from utils.general import LOGGER

try:
    import thop  # for FLOPs computation
except ImportError:
    thop = None


def date_modified(path=__file__):
    # return human-readable file modification date, i.e. '2021-3-26'
    t = datetime.datetime.fromtimestamp(Path(path).stat().st_mtime)
    return f'{t.year}-{t.month}-{t.day}'


def git_describe(path=Path(__file__).parent):  # path must be a directory
    # return human-readable git description, i.e. v5.0-5-g3e25f1e https://git-scm.com/docs/git-describe
    s = f'git -C {path} describe --tags --long --always'
    try:
        return subprocess.check_output(s, shell=True, stderr=subprocess.STDOUT).decode()[:-1]
    except subprocess.CalledProcessError as e:
        return ''  # not a git repository


# def select_device(device='', batch_size=0, newline=True):
#     # device = 'cpu' or '0' or '0,1,2,3'
#     s = f'YOLOv5 🚀 {git_describe() or date_modified()} torch {torch.__version__} '  # string
#     device = str(device).strip().lower().replace('cuda:', '')  # to string, 'cuda:0' to '0'
#     cpu = device == 'cpu'
#     if cpu:
#         os.environ['CUDA_VISIBLE_DEVICES'] = '-1'  # force torch.cuda.is_available() = False
#     elif device:  # non-cpu device requested
#         os.environ['CUDA_VISIBLE_DEVICES'] = device  # set environment variable
#         assert torch.cuda.is_available(), f'CUDA unavailable, invalid device {device} requested'  # check availability

#     cuda = not cpu and torch.cuda.is_available()
#     if cuda:
#         devices = device.split(',') if device else '0'  # range(torch.cuda.device_count())  # i.e. 0,1,6,7
#         n = len(devices)  # device count
#         if n > 1 and batch_size > 0:  # check batch_size is divisible by device_count
#             assert batch_size % n == 0, f'batch-size {batch_size} not multiple of GPU count {n}'
#         space = ' ' * (len(s) + 1)
#         for i, d in enumerate(devices):
#             p = torch.cuda.get_device_properties(i)
#             s += f"{'' if i == 0 else space}CUDA:{d} ({p.name}, {p.total_memory / 1024 ** 2:.0f}MiB)\n"  # bytes to MB
#     else:
#         s += 'CPU\n'

#     if not newline:
#         s = s.rstrip()
#     # LOGGER.info(s.encode().decode('ascii', 'ignore') if platform.system() == 'Windows' else s)  # emoji-safe
#     return torch.device('cuda:0' if cuda else 'cpu')

def select_device(device='', batch_size=0, newline=True):
    # device = 'cpu' or '0' or '0,1,2,3'
    # s = f'YOLOv5 🚀 {git_describe() or date_modified()} torch {torch.__version__} '  # string
    device = str(device).strip().lower().replace(
        'cuda:', '')  # to string, 'cuda:0' to '0'
    cpu = device == 'cpu'
    if cpu:
        # force torch.cuda.is_available() = False
        os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
    elif device:  # non-cpu device requested
        pass
        ######## comment 
        # os.environ['CUDA_VISIBLE_DEVICES'] = device  # set environment variable
        assert torch.cuda.is_available(
        ), f'CUDA unavailable, invalid device {device} requested'  # check availability

    cuda = not cpu and torch.cuda.is_available()
    if cuda:
        # range(torch.cuda.device_count())  # i.e. 0,1,6,7
        devices = device.split(',') if device else ['0']
        n = len(devices)  # device count
        if n > 1 and batch_size > 0:  # check batch_size is divisible by device_count
            assert batch_size % n == 0, f'batch-size {batch_size} not multiple of GPU count {n}'
        # space = ' ' * (len(s) + 1)
        # for i, d in enumerate(devices):
            # p = torch.cuda.get_device_properties(i)
            # s += f"{'' if i == 0 else space}CUDA:{d} ({p.name}, {p.total_memory / 1024 ** 2:.0f}MiB)\n"  # bytes to MB
    # else:
        # s += 'CPU\n'

    # if not newline:
    #     s = s.rstrip()
    return torch.device('cuda:{}'.format(devices[0]) if cuda else 'cpu')

def time_sync():
    # pytorch-accurate time
    if torch.cuda.is_available():
        torch.cuda.synchronize()
    return time.time()


def profile(input, ops, n=10, device=None):
    # YOLOv5 speed/memory/FLOPs profiler
    #
    # Usage:
    #     input = torch.randn(16, 3, 640, 640)
    #     m1 = lambda x: x * torch.sigmoid(x)
    #     m2 = nn.SiLU()
    #     profile(input, [m1, m2], n=100)  # profile over 100 iterations

    results = []
    device = device or select_device()
    print(f"{'Params':>12s}{'GFLOPs':>12s}{'GPU_mem (GB)':>14s}{'forward (ms)':>14s}{'backward (ms)':>14s}"
          f"{'input':>24s}{'output':>24s}")

    for x in input if isinstance(input, list) else [input]:
        x = x.to(device)
        x.requires_grad = True
        for m in ops if isinstance(ops, list) else [ops]:
            m = m.to(device) if hasattr(m, 'to') else m  # device
            m = m.half() if hasattr(m, 'half') and isinstance(x, torch.Tensor) and x.dtype is torch.float16 else m
            tf, tb, t = 0, 0, [0, 0, 0]  # dt forward, backward
            try:
                flops = thop.profile(m, inputs=(x,), verbose=False)[0] / 1E9 * 2  # GFLOPs
            except:
                flops = 0

            try:
                for _ in range(n):
                    t[0] = time_sync()
                    y = m(x)
                    t[1] = time_sync()
                    try:
                        _ = (sum(yi.sum() for yi in y) if isinstance(y, list) else y).sum().backward()
                        t[2] = time_sync()
                    except Exception as e:  # no backward method
                        # print(e)  # for debug
                        t[2] = float('nan')
                    tf += (t[1] - t[0]) * 1000 / n  # ms per op forward
                    tb += (t[2] - t[1]) * 1000 / n  # ms per op backward
                mem = torch.cuda.memory_reserved() / 1E9 if torch.cuda.is_available() else 0  # (GB)
                s_in = tuple(x.shape) if isinstance(x, torch.Tensor) else 'list'
                s_out = tuple(y.shape) if isinstance(y, torch.Tensor) else 'list'
                p = sum(list(x.numel() for x in m.parameters())) if isinstance(m, nn.Module) else 0  # parameters
                print(f'{p:12}{flops:12.4g}{mem:>14.3f}{tf:14.4g}{tb:14.4g}{str(s_in):>24s}{str(s_out):>24s}')
                results.append([p, flops, mem, tf, tb, s_in, s_out])
            except Exception as e:
                print(e)
                results.append(None)
            torch.cuda.empty_cache()
    return results


def is_parallel(model):
    # Returns True if model is of type DP or DDP
    return type(model) in (nn.parallel.DataParallel, nn.parallel.DistributedDataParallel)

def initialize_weights(model):
    for m in model.modules():
        t = type(m)
        if t is nn.Conv2d:
            pass  # nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
        elif t is nn.BatchNorm2d:
            m.eps = 1e-3
            m.momentum = 0.03
        elif t in [nn.Hardswish, nn.LeakyReLU, nn.ReLU, nn.ReLU6, nn.SiLU]:
            m.inplace = True


def find_modules(model, mclass=nn.Conv2d):
    # Finds layer indices matching module class 'mclass'
    return [i for i, m in enumerate(model.module_list) if isinstance(m, mclass)]


def sparsity(model):
    # Return global model sparsity
    a, b = 0, 0
    for p in model.parameters():
        a += p.numel()
        b += (p == 0).sum()
    return b / a


def prune(model, amount=0.3):
    # Prune model to requested global sparsity
    import torch.nn.utils.prune as prune
    print('Pruning model... ', end='')
    for name, m in model.named_modules():
        if isinstance(m, nn.Conv2d):
            prune.l1_unstructured(m, name='weight', amount=amount)  # prune
            prune.remove(m, 'weight')  # make permanent
    print(' %.3g global sparsity' % sparsity(model))


def fuse_conv_and_bn(conv, bn):
    # Fuse convolution and batchnorm layers https://tehnokv.com/posts/fusing-batchnorm-and-conv/
    fusedconv = nn.Conv2d(conv.in_channels,
                          conv.out_channels,
                          kernel_size=conv.kernel_size,
                          stride=conv.stride,
                          padding=conv.padding,
                          groups=conv.groups,
                          bias=True).requires_grad_(False).to(conv.weight.device)

    # prepare filters
    w_conv = conv.weight.clone().view(conv.out_channels, -1)
    w_bn = torch.diag(bn.weight.div(torch.sqrt(bn.eps + bn.running_var)))
    fusedconv.weight.copy_(torch.mm(w_bn, w_conv).view(fusedconv.weight.shape))

    # prepare spatial bias
    b_conv = torch.zeros(conv.weight.size(0), device=conv.weight.device) if conv.bias is None else conv.bias
    b_bn = bn.bias - bn.weight.mul(bn.running_mean).div(torch.sqrt(bn.running_var + bn.eps))
    fusedconv.bias.copy_(torch.mm(w_bn, b_conv.reshape(-1, 1)).reshape(-1) + b_bn)

    return fusedconv


def model_info(model, verbose=False, img_size=640):
    # Model information. img_size may be int or list, i.e. img_size=640 or img_size=[640, 320]
    n_p = sum(x.numel() for x in model.parameters())  # number parameters
    n_g = sum(x.numel() for x in model.parameters() if x.requires_grad)  # number gradients
    if verbose:
        print(f"{'layer':>5} {'name':>40} {'gradient':>9} {'parameters':>12} {'shape':>20} {'mu':>10} {'sigma':>10}")
        for i, (name, p) in enumerate(model.named_parameters()):
            name = name.replace('module_list.', '')
            print('%5g %40s %9s %12g %20s %10.3g %10.3g' %
                  (i, name, p.requires_grad, p.numel(), list(p.shape), p.mean(), p.std()))

    try:  # FLOPs
        from thop import profile
        stride = max(int(model.stride.max()), 32) if hasattr(model, 'stride') else 32
        img = torch.zeros((1, model.yaml.get('ch', 3), stride, stride), device=next(model.parameters()).device)  # input
        flops = profile(deepcopy(model), inputs=(img,), verbose=False)[0] / 1E9 * 2  # stride GFLOPs
        img_size = img_size if isinstance(img_size, list) else [img_size, img_size]  # expand if int/float
        fs = ', %.1f GFLOPs' % (flops * img_size[0] / stride * img_size[1] / stride)  # 640x640 GFLOPs
    except (ImportError, Exception):
        fs = ''

    # LOGGER.info(f"Model Summary: {len(list(model.modules()))} layers, {n_p} parameters, {n_g} gradients{fs}")


def scale_img(img, ratio=1.0, same_shape=False, gs=32):  # img(16,3,256,416)
    # scales img(bs,3,y,x) by ratio constrained to gs-multiple
    if ratio == 1.0:
        return img
    else:
        h, w = img.shape[2:]
        s = (int(h * ratio), int(w * ratio))  # new size
        img = F.interpolate(img, size=s, mode='bilinear', align_corners=False)  # resize
        if not same_shape:  # pad/crop img
            h, w = (math.ceil(x * ratio / gs) * gs for x in (h, w))
        return F.pad(img, [0, w - s[1], 0, h - s[0]], value=0.447)  # value = imagenet mean





from .embedding import Embeddings, LearnPositionalEncoding, PositionalEncoding, TokenTypeEncoding
from .multi_headed_attn import MultiHeadedAttention, ATTEN_TYPE_SELF, ATTEN_TYPE_HYGRA
from .util_class import LayerNorm, PositionwiseFeedForward, ACTIVATION_FUNCTIONS, ActivationFunction, RMSNorm, MLP
from .misc import sequence_mask, aeq
from .average_attn import AverageAttention
from .embedding import Embeddings, PADDING_INDEX, _position_encoding_default, _position_encoding_absolute
from .encoder import TransformerEncoder
from .decoder import TransformerDecoder
from .global_attention import GlobalAttention
from .base_model import EncoderBase, DecoderBase
from .attention import DotProductAttention, EmbeddingAttentionLayer


"""
Callback utils
"""


class Callbacks:
    """"
    Handles all registered callbacks for train Hooks
    Usage:
        class Loggers():
            def on_train_start(self):
                pass
            ...
        loggers = Loggers()
        # Register actions
        for k in methods(loggers):
            callbacks.register_action(k, callback=getattr(loggers, k))
            
example usage
train.py:
   68:     callbacks.run('on_pretrain_routine_start')
  229:     callbacks.run('on_pretrain_routine_end')
  258:     callbacks.run('on_train_start')
  264:     callbacks.run('on_train_epoch_start')
  286:     callbacks.run('on_train_batch_start')
  338:     callbacks.run('on_train_batch_end', ni, model, imgs, targets, paths, plots)
  349:     callbacks.run('on_train_epoch_end', epoch=epoch)
  371:     callbacks.run('on_fit_epoch_end', log_vals, epoch, best_fitness, fi)
  392:     callbacks.run('on_model_save', last, epoch, final_epoch, best_fitness, fi)
  427:     callbacks.run('on_fit_epoch_end', list(mloss) + list(results) + lr, epoch, best_fitness, fi)
  429:     callbacks.run('on_train_end', last, best, plots, epoch, results)

val.py:
  191:     callbacks.run('on_val_start')
  194:     callbacks.run('on_val_batch_start')
  256:     callbacks.run('on_val_image_end', pred, predn, path, names, im[si])
  263:     callbacks.run('on_val_batch_end')
  293:     callbacks.run('on_val_end')

    """

    def __init__(self):
        # Define the available callbacks
        self._callbacks = {
            'on_pretrain_routine_start': [],
            'on_pretrain_routine_end': [],
            'on_train_start': [],
            'on_train_epoch_start': [],
            'on_train_batch_start': [],
            'optimizer_step': [],
            'on_before_zero_grad': [],
            'on_train_batch_end': [],
            'on_train_epoch_end': [],
            'on_val_start': [],
            'on_val_batch_start': [],
            'on_val_image_end': [],
            'on_val_batch_end': [],
            'on_val_end': [],
            'on_fit_epoch_end': [],  # fit = train + val
            'on_model_save': [],
            'on_train_end': [],
            'on_params_update': [],
            'teardown': [], }
        self.stop_training = False  # set True to interrupt training

    def register_action(self, hook, name='', callback=None):
        """
        Register a new action to a callback hook

        Args:
            hook: The callback hook name to register the action to
            name: The name of the action for later reference
            callback: The callback to fire
        """
        assert hook in self._callbacks, f"hook '{hook}' not found in callbacks {self._callbacks}"
        assert callable(callback), f"callback '{callback}' is not callable"
        self._callbacks[hook].append({'name': name, 'callback': callback})

    def get_registered_actions(self, hook=None):
        """"
        Returns all the registered actions by callback hook

        Args:
            hook: The name of the hook to check, defaults to all
        """
        return self._callbacks[hook] if hook else self._callbacks

    def run(self, hook, *args, **kwargs):
        """
        Loop through the registered actions and fire all callbacks

        Args:
            hook: The name of the hook to check, defaults to all
            args: Arguments to receive from YOLOv5
            kwargs: Keyword Arguments to receive from YOLOv5
        """

        assert hook in self._callbacks, f"hook '{hook}' not found in callbacks {self._callbacks}"

        for logger in self._callbacks[hook]:
            logger['callback'](*args, **kwargs)


    def run_once(self, hook, *args, **kwargs):
        assert hook in self._callbacks, f"hook '{hook}' not found in callbacks {self._callbacks}"

        for logger in self._callbacks[hook]:
            return logger['callback'](*args, **kwargs) 



# from multiprocessing import Process
# import multiprocessing
import aicmder as cmder
import json
from aicmder.service.PPpolicy import *
from random import randint
import time
import zmq
import datetime, json, sys, os
from termcolor import colored
from aicmder.service import WORKER_PORT
import signal


def handler(signum, frame):
    raise Exception('Action took too much time')


def worker_socket(context, poller):
    """Helper function that returns a new configured socket
       connected to the Paranoid Pirate queue"""
    worker = context.socket(zmq.DEALER) # DEALER
    identity = b"%04X-%04X" % (randint(0, 0x10000), randint(0, 0x10000))
    worker.setsockopt(zmq.IDENTITY, identity)
    poller.register(worker, zmq.POLLIN)
    worker.connect(f"tcp://localhost:{WORKER_PORT}")
    worker.send(PPP_READY)
    return worker

ModuleName = cmder.ModuleDefine.ModuleName.str()
ModuleMethod = cmder.ModuleDefine.ModuleMethod.str()
ModuleInitArgs = cmder.ModuleDefine.ModuleInitArgs.str()
ModuleParams = cmder.ModuleDefine.ModuleParams.str()
DeviceId = cmder.ModuleDefine.DeviceId.str()
WorkerId = cmder.ModuleDefine.WorkerId.str()
############## important timeout
REQUEST_TIMEOUT = 20 * 100   # timeout 20 seconds * 100


class Worker:
    
    
    def __init__(self, worker_id, module_infos, logger, device_id=-1, **kwargs):
        super(Worker, self).__init__()
        if type(module_infos) == dict:
            self.module_infos = module_infos
        else:
            self.module_infos = json.loads(module_infos) if type(module_infos) == str else module_infos
        self.device_id = device_id
        self.kwargs = kwargs
        
        assert len(self.module_infos) > 0
        self.serving_methods = {}
        self.logger = logger
        self.module_name = None
        for module_name, module_info in self.module_infos.items():
            self.logger.info(module_name, module_info)
            
            init_args = module_info.get(ModuleInitArgs, {})
            init_args.update({ModuleName: module_info[ModuleName], DeviceId: self.device_id, WorkerId: worker_id})
            init_args.update(module_info[ModuleInitArgs])
            
            # module = cmder.Module(**init_args)
            module = cmder.Module.new_module(**init_args)

            
            method_name = module.serving_func_name
            serving_method = getattr(module, method_name)
            # serving_args = module_info.get(ModuleParams, {})
            self.serving_methods[module_name] = {ModuleName: module_info[ModuleName], ModuleMethod: serving_method}
            if self.module_name is None:
                self.module_name = module_name
        # print(self.serving_methods)
        # self.is_ready = multiprocessing.Event()
        
        
    def run(self):    
        context = zmq.Context(1)
        poller = zmq.Poller()

        liveness = HEARTBEAT_LIVENESS
        interval = INTERVAL_INIT
        # self.is_ready.set()
        heartbeat_at = time.time() + HEARTBEAT_INTERVAL
        self.logger.info("worker is ready!")
        worker = worker_socket(context, poller)
        while True:
            socks = dict(poller.poll(HEARTBEAT_INTERVAL * 1000))

            # Handle worker activity on backend
            if socks.get(worker) == zmq.POLLIN:
                #  Get message
                #  - 3-part envelope + content -> request
                #  - 1-part HEARTBEAT -> heartbeat
                frames = worker.recv_multipart()
                if not frames:
                    break # Interrupted

                if len(frames) == 3:
                    module = self.serving_methods[self.module_name]
                    serving_args = frames[len(frames) - 1].decode()
                    try:
                        req_module_name = serving_args.partition('\n')[0]
                        if len(req_module_name) > 0:
                            if req_module_name in self.serving_methods:
                                 module = self.serving_methods[req_module_name]
                        # else:
                        #     print("no module name")
                        serving_args = serving_args.split("\n", 1)[1]
                        serving_args_dict = json.loads(serving_args)
                    except Exception as e:
                        # just in case if not json
                        serving_args_dict = serving_args

                    signal.signal(signal.SIGALRM, handler)
                    signal.alarm(REQUEST_TIMEOUT) #Set the parameter to the amount of seconds you want to wait
                    try:
                        if type(serving_args_dict) is dict:
                            frames[len(frames) - 1] = module[ModuleMethod](**serving_args_dict).encode()
                        else:
                            frames[len(frames) - 1] = module[ModuleMethod](serving_args_dict).encode()
                    except Exception as e:
                        self.logger.info(e)
                        frames[len(frames) - 1] = '{"code":-1}'.encode()
                    signal.alarm(0) #Disables the alarm 

                    self.logger.debug("I: Normal reply")
                    worker.send_multipart(frames)
                    liveness = HEARTBEAT_LIVENESS
                    
                elif len(frames) == 1 and frames[0] == PPP_HEARTBEAT:
                    # self.logger.debug("I: Queue heartbeat", datetime.datetime.now())
                    liveness = HEARTBEAT_LIVENESS
                else:
                    self.logger.debug("E: Invalid message: %s" % frames)
                interval = INTERVAL_INIT
            else:
                liveness -= 1
                if liveness == 0:
                    self.logger.info("W: Heartbeat failure, can't reach queue")
                    self.logger.info("W: Reconnecting in %0.2fs..." % interval)
                    time.sleep(interval)

                    if interval < INTERVAL_MAX:
                        interval *= 2
                    poller.unregister(worker)
                    worker.setsockopt(zmq.LINGER, 0)
                    worker.close()
                    worker = worker_socket(context, poller)
                    liveness = HEARTBEAT_LIVENESS
            if time.time() > heartbeat_at:
                heartbeat_at = time.time() + HEARTBEAT_INTERVAL
                # self.logger.debug("I: Worker heartbeat", datetime.datetime.now())
                worker.send(PPP_HEARTBEAT)
                
if __name__ == "__main__":
    # config = {'albert': {'name': 'tests_model/AlbertModule', 'init_args': {'dummpy_params': 'dummpy'}}}
    argv = sys.argv
    assert len(argv) >= 3
    config = json.loads(argv[1])
    device_id = int(argv[2])
    worker_idx = int(argv[3])   
    
    worker_id = os.getpid()
    verbose = cmder.Common.LOG_VERBOSE
    logger = cmder.Common.set_logger(colored('WORKER-%d' % worker_id, 'yellow'), verbose)
        
    worker = Worker(worker_idx, config, logger, device_id = device_id)
    worker.run()


import onnxruntime as ort
import numpy as np
from yolo_util import *
import cv2
import time

# options = self.sess.get_provider_options()
# print(options)
# cuda_options = options['CUDAExecutionProvider']
# cuda_options['cudnn_conv_use_max_workspace'] = '1'
# self.sess.set_providers(['CUDAExecutionProvider'], [cuda_options])

# so = ort.SessionOptions()
# so.execution_mode = ort.ExecutionMode.ORT_SEQUENTIAL
# so.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_ALL

# exproviders = ['CUDAExecutionProvider', 'CPUExecutionProvider']
# self.sess = ort.InferenceSession('/home/faith/aicmder/tests_model/best.onnx', so, providers=exproviders)

# if "cuda" in device:
#     providers = ["CUDAExecutionProvider"]
# else:
#     providers = ["CPUExecutionProvider"]


class Yolov5:

    def __init__(self, imgsz, model='/home/faith/yolov5m6.onnx', names=[''], classes=None, conf_thres=0.55, iou_thres=0.45, fp16=False) -> None:
        # python export.py --weights runs/train/exp26/weights/best.pt --img 736 1280
        print(ort.get_available_providers())

        cuda = False
        providers = ['CUDAExecutionProvider', 'CPUExecutionProvider'] if cuda else [
            'CPUExecutionProvider']

        # https://stackoverflow.com/questions/70844974/onnxruntime-vs-onnxruntimeopenvinoep-inference-time-difference
        # pip install onnxruntime-openvino
        device = 'CPU_FP32'
        self.sess = ort.InferenceSession(model, providers=['OpenVINOExecutionProvider'], provider_options=[{'device_type' : device}])

        # self.sess = ort.InferenceSession(model, providers=providers)
        # self.sess.set_providers(['CUDAExecutionProvider'], [ {'device_id': 1}])

        print(ort.get_device(), providers, "----")
        self.stride = 64
        self.hide_labels, self.hide_conf = False, False
        # self.names = ['enermy']
        self.names = names
        self.fp16 = fp16
        self.imgsz = imgsz  # (768, 1280)
        self.conf_thres, self.iou_thres, self.classes, self.agnostic_nms, self.max_det = conf_thres, iou_thres, classes, False, 1000

    def non_max_suppression(self, prediction, input_h, input_w, origin_h, origin_w, conf_thres, nms_thres, classes=None):
        """
        description: Removes detections with lower object confidence score than 'conf_thres' and performs
        Non-Maximum Suppression to further filter detections.
        param:
            prediction: detections, (x1, y1, x2, y2, conf, cls_id)
            origin_h: original image height
            origin_w: original image width
            conf_thres: a confidence threshold to filter detections
            nms_thres: a iou threshold to filter detections
        return:
            boxes: output after nms with the shape (x1, y1, x2, y2, conf, cls_id)
        """
        nc = prediction.shape[2] - 5
        multi_label = False
        multi_label &= nc > 1  # multiple labels per box (adds 0.5ms/img)

        mask = prediction[..., 4] >= conf_thres

        # Get the boxes that score > CONF_THRESH
        print(prediction.shape)
        boxes = prediction[mask]

        # Trandform bbox from [center_x, center_y, w, h] to [x1, y1, x2, y2]
        boxes[:, :4] = self.xywh2xyxy(
            input_h, input_w, origin_h, origin_w, boxes[:, :4])
        # clip the coordinates
        boxes[:, 0] = np.clip(boxes[:, 0], 0, origin_w - 1)
        boxes[:, 2] = np.clip(boxes[:, 2], 0, origin_w - 1)
        boxes[:, 1] = np.clip(boxes[:, 1], 0, origin_h - 1)
        boxes[:, 3] = np.clip(boxes[:, 3], 0, origin_h - 1)
        # Object confidence
        confs = boxes[:, 4]

        boxes[:, 5:] *= boxes[:, 4:5]  # conf = obj_conf * cls_conf
        if multi_label:
            assert "not implemented"
        else:
            # output_shape = boxes[:, 5:].shape
            j = np.argmax(boxes[:, 5:], axis=1).reshape(-1, 1)
            conf = np.max(boxes[:, 5:], axis=1).reshape(-1, 1)
            boxes = np.concatenate((boxes[:, :4], conf, j), 1)
            # conf, j = boxes[:, 5:].max(1, keepdim=True)

        # Filter by class
        if classes is not None:
            class_selection = (boxes[:, 5:6] == classes).any(1)
            boxes = boxes[class_selection]
            confs = confs[class_selection]

        ################################################
        # Sort by the confs
        boxes = boxes[np.argsort(-confs)]
        # Perform non-maximum suppression
        keep_boxes = []
        while boxes.shape[0]:
            large_overlap = bbox_iou(np.expand_dims(
                boxes[0, :4], 0), boxes[:, :4]) > nms_thres
            label_match = boxes[0, -1] == boxes[:, -1]
            # Indices of boxes with lower confidence scores, large IOUs and matching labels
            invalid = large_overlap & label_match
            keep_boxes += [boxes[0]]
            boxes = boxes[~invalid]
        boxes = np.stack(keep_boxes, 0) if len(keep_boxes) else np.array([])
        ################################################

        scores = boxes[:, 4]
        # ii = torch.ops.torchvision.nms(torch.from_numpy(boxes[:, :4]), torch.from_numpy(scores), nms_thres)
        i = numpy_nms(boxes[:, :4], scores, nms_thres)
        boxes = boxes[i]

        return boxes

    def xywh2xyxy(self, input_h, input_w, origin_h, origin_w, x):
        """
        description:    Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
        param:
            origin_h:   height of original image
            origin_w:   width of original image
            x:          A boxes numpy, each row is a box [center_x, center_y, w, h]
        return:
            y:          A boxes numpy, each row is a box [x1, y1, x2, y2]
        """
        y = np.zeros_like(x)
        r_w = input_w / origin_w
        r_h = input_h / origin_h
        if r_h > r_w:
            y[:, 0] = x[:, 0] - x[:, 2] / 2
            y[:, 2] = x[:, 0] + x[:, 2] / 2
            y[:, 1] = x[:, 1] - x[:, 3] / 2 - (input_h - r_w * origin_h) / 2
            y[:, 3] = x[:, 1] + x[:, 3] / 2 - (input_h - r_w * origin_h) / 2
            y /= r_w
        else:
            y[:, 0] = x[:, 0] - x[:, 2] / 2 - (input_w - r_h * origin_w) / 2
            y[:, 2] = x[:, 0] + x[:, 2] / 2 - (input_w - r_h * origin_w) / 2
            y[:, 1] = x[:, 1] - x[:, 3] / 2
            y[:, 3] = x[:, 1] + x[:, 3] / 2
            y /= r_h

        return y

    def predict_image(self, img_bgr, debug=0):
        # img_bgr = cv2.resize(img_bgr, self.imgsz)
        img = letterbox(img_bgr, self.imgsz, stride=self.stride)[0]
        print(img.shape, img_bgr.shape)
        img = img.transpose((2, 0, 1))[::-1]  # HWC to CHW, BGR to RGB
        img = np.ascontiguousarray(img).astype(np.float32)
        img /= 255.0  # 0 - 255 to 0.0 - 1.0
        if len(img.shape) == 3:
            img = img[None]  # expand for batch dim
        start = time.time()
        pred = self.sess.run([self.sess.get_outputs()[0].name], {
                             self.sess.get_inputs()[0].name: img.astype(np.float16) if self.fp16 else img})[0]
        end = time.time()
        print("time--", end - start)

        h, w, _ = img_bgr.shape
        # start = time.time()
        origin_h, origin_w, _ = img_bgr.shape
        input_h, input_w = img.shape[2:]
        # # print(origin_h, origin_w, input_h, input_w)
        boxes = self.non_max_suppression(pred, input_h, input_w, origin_h, origin_w,
                                         conf_thres=self.conf_thres, nms_thres=self.iou_thres, classes=self.classes)
        # end = time.time()
        # print("new time---", boxes.shape, end - start, self.conf_thres, self.iou_thres)
        # # print(boxes)

        resp_list = []
        for i, det in enumerate(boxes):
            if len(det) == 6:
                # print(det)
                start_x, start_y, end_x, end_y, conf, cls = det

                c = int(cls)  # integer class
                label = None if self.hide_labels else (
                    self.names[c] if self.hide_conf else f'{self.names[c]} {conf:.2f}')
                # start_x, start_y, end_x, end_y = int(xyxy[0]), int(xyxy[1]), int(xyxy[2]), int(xyxy[3])
                resp_list.append({
                    "start_x": int(start_x),
                    "start_y": int(start_y),
                    "end_x": int(end_x),
                    "end_y": int(end_y),
                    "x0": start_x / w,
                    "x1": end_x / w,
                    "y0": start_y / h,
                    "y1": end_y / h,
                    "c": c,
                    "label": label,
                    "conf": float(conf)
                })
                if debug > 0:
                    plot_one_box((start_x, start_y, end_x, end_y),
                                 img_bgr, label=label)

        # pred = non_max_suppression(pred, self.conf_thres, self.iou_thres, self.classes, self.agnostic_nms, max_det=self.max_det)
        # for i, det in enumerate(pred):  # detections per image
        #     if len(det) > 0:
        #         det[:, :4] = scale_coords(img.shape[2:], det[:, :4], img_bgr.shape).round()
        #         # print("--", det)
        #         for *xyxy, conf in reversed(det):
        #             if len(xyxy) >= 4:
        #                 # print(xyxy, conf)
        #                 plot_one_box(xyxy, img_bgr, label=self.names)
        if debug > 1:
            cv2.imwrite('detect.png', img_bgr)


def onnx_convert():
    from onnxconverter_common.float16 import convert_float_to_float16
    import onnx
    model = onnx.load_model('/home/faith/yolov5m6.onnx')
    fp16_model = convert_float_to_float16(model)
    onnx.save_model(fp16_model, '/home/faith/yolov5m6_fp16.onnx')


def inference():
    model_name = '/home/faith/yolov5m6.onnx'
    # model_name = '/home/faith/yolov5m6_fp16.onnx'

    yolo = Yolov5(model=model_name, names=['person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus', 'train', 'truck', 'boat', 'traffic light', 'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird', 'cat', 'dog', 'horse', 'sheep', 'cow', 'elephant', 'bear', 'zebra', 'giraffe', 'backpack', 'umbrella', 'handbag', 'tie', 'suitcase', 'frisbee', 'skis', 'snowboard', 'sports ball', 'kite', 'baseball bat', 'baseball glove', 'skateboard', 'surfboard',
                                                            'tennis racket', 'bottle', 'wine glass', 'cup', 'fork', 'knife', 'spoon', 'bowl', 'banana', 'apple', 'sandwich', 'orange', 'broccoli', 'carrot', 'hot dog', 'pizza', 'donut', 'cake', 'chair', 'couch', 'potted plant', 'bed', 'dining table', 'toilet', 'tv', 'laptop', 'mouse', 'remote', 'keyboard', 'cell phone', 'microwave', 'oven', 'toaster', 'sink', 'refrigerator', 'book', 'clock', 'vase', 'scissors', 'teddy bear', 'hair drier', 'toothbrush'], classes=[0], imgsz=(640, 640), fp16=False)
    image_file = "/home/faith/aicmder/tests_model/222.jpg"
    image_file = "/home/faith/aicmder/dl/IMG_5980.jpg"
    for i in range(5):
        img = cv2.imread(image_file, cv2.COLOR_RGB2BGR)
        # time.sleep(2)
        start = time.time()
        yolo.predict_image(img, debug=0)
        end = time.time()
        print(end - start)


if __name__ == "__main__":
    # onnx_convert()
    inference()



# src: https://github.com/OpenNMT/OpenNMT-py/blob/master/onmt/decoders/decoder.py


class DecoderState(object):
    """Interface for grouping together the current state of a recurrent
    decoder. In the simplest case just represents the hidden state of
    the model.  But can also be used for implementing various forms of
    input_feeding and non-recurrent models.
    Modules need to implement this to utilize beam search decoding.
    """

    def detach(self):
        """ Need to document this """
        self.hidden = tuple([_.detach() for _ in self.hidden])

    def beam_update(self, idx, positions, beam_size):
        """ Need to document this """
        for e in self._all:
            sizes = e.size()
            br = sizes[1]
            if len(sizes) == 3:
                sent_states = e.view(sizes[0], beam_size, br // beam_size,
                                     sizes[2])[:, :, idx]
            else:
                sent_states = e.view(sizes[0], beam_size,
                                     br // beam_size,
                                     sizes[2],
                                     sizes[3])[:, :, idx]

            sent_states.data.copy_(
                sent_states.data.index_select(1, positions))

    def map_batch_fn(self, fn):
        raise NotImplementedError()


class RNNDecoderState(DecoderState):
    """ Base class for RNN decoder state """

    def __init__(self, hidden_size, rnnstate):
        """
        Args:
            hidden_size (int): the size of hidden layer of the decoder.
            rnnstate: final hidden state from the encoder.
                transformed to shape: layers x batch x (directions*dim).
        """
        if not isinstance(rnnstate, tuple):
            self.hidden = (rnnstate,)
        else:
            self.hidden = rnnstate
        self.coverage = None

    @property
    def _all(self):
        return self.hidden

    def update_state(self, rnnstate, coverage):
        """ Update decoder state """
        if not isinstance(rnnstate, tuple):
            self.hidden = (rnnstate,)
        else:
            self.hidden = rnnstate
        self.coverage = coverage

    def repeat_beam_size_times(self, beam_size):
        """ Repeat beam_size times along batch dimension. """
        vars = [e.data.repeat(1, beam_size, 1)
                for e in self._all]
        self.hidden = tuple(vars)

    def map_batch_fn(self, fn):
        self.hidden = tuple(map(lambda x: fn(x, 1), self.hidden))



import pickle
from copy import deepcopy
from thop import profile
import torch
import pickle
import os
from aicmder.utils.general import try_except
import gc
import datetime
import inspect

import numpy as np

dtype_memory_size_dict = {
    torch.float64: 64/8,
    torch.double: 64/8,
    torch.float32: 32/8,
    torch.float: 32/8,
    torch.float16: 16/8,
    torch.half: 16/8,
    torch.int64: 64/8,
    torch.long: 64/8,
    torch.int32: 32/8,
    torch.int: 32/8,
    torch.int16: 16/8,
    torch.short: 16/6,
    torch.uint8: 8/8,
    torch.int8: 8/8,
}
# compatibility of torch1.0
if getattr(torch, "bfloat16", None) is not None:
    dtype_memory_size_dict[torch.bfloat16] = 16/8
if getattr(torch, "bool", None) is not None:
    dtype_memory_size_dict[torch.bool] = 8/8  # pytorch use 1 byte for a bool, see https://github.com/pytorch/pytorch/issues/41571


def get_mem_space(x):
    try:
        ret = dtype_memory_size_dict[x]
    except KeyError:
        print(f"dtype {x} is not supported!")
    return ret


class MemTracker(object):
    """
    Class used to track pytorch memory usage
    Arguments:
        detail(bool, default True): whether the function shows the detail gpu memory usage
        path(str): where to save log file
        verbose(bool, default False): whether show the trivial exception
        device(int): GPU number, default is 0
    """

    def __init__(self, detail=True, path='', verbose=False, device=0):
        self.print_detail = detail
        self.last_tensor_sizes = set()
        self.gpu_profile_fn = path + f'{datetime.datetime.now():%d-%b-%y-%H:%M:%S}-gpu_mem_track.txt'
        self.verbose = verbose
        self.begin = True
        self.device = device

    def get_tensors(self):
        for obj in gc.get_objects():
            try:
                if torch.is_tensor(obj) or (hasattr(obj, 'data') and torch.is_tensor(obj.data)):
                    tensor = obj
                else:
                    continue
                if tensor.is_cuda:
                    yield tensor
            except Exception as e:
                if self.verbose:
                    print('A trivial exception occured: {}'.format(e))

    def get_tensor_usage(self):
        sizes = [np.prod(np.array(tensor.size())) * get_mem_space(tensor.dtype) for tensor in self.get_tensors()]
        return np.sum(sizes) / 1024**2

    def get_allocate_usage(self):
        return torch.cuda.memory_allocated() / 1024**2

    def clear_cache(self):
        gc.collect()
        torch.cuda.empty_cache()

    def print_all_gpu_tensor(self, file=None):
        for x in self.get_tensors():
            print(x.size(), x.dtype, np.prod(np.array(x.size()))*get_mem_space(x.dtype)/1024**2, file=file)

    def track(self):
        """
        Track the GPU memory usage
        """
        frameinfo = inspect.stack()[1]
        where_str = frameinfo.filename + ' line ' + str(frameinfo.lineno) + ': ' + frameinfo.function

        with open(self.gpu_profile_fn, 'a+') as f:

            if self.begin:
                f.write(f"GPU Memory Track | {datetime.datetime.now():%d-%b-%y-%H:%M:%S} |"
                        f" Total Tensor Used Memory:{self.get_tensor_usage():<7.1f}Mb"
                        f" Total Allocated Memory:{self.get_allocate_usage():<7.1f}Mb\n\n")
                self.begin = False

            if self.print_detail is True:
                ts_list = [(tensor.size(), tensor.dtype) for tensor in self.get_tensors()]
                new_tensor_sizes = {(type(x),
                                     tuple(x.size()),
                                     ts_list.count((x.size(), x.dtype)),
                                     np.prod(np.array(x.size()))*get_mem_space(x.dtype)/1024**2,
                                     x.dtype) for x in self.get_tensors()}
                for t, s, n, m, data_type in new_tensor_sizes - self.last_tensor_sizes:
                    f.write(f'+ | {str(n)} * Size:{str(s):<20} | Memory: {str(m*n)[:6]} M | {str(t):<20} | {data_type}\n')
                for t, s, n, m, data_type in self.last_tensor_sizes - new_tensor_sizes:
                    f.write(f'- | {str(n)} * Size:{str(s):<20} | Memory: {str(m*n)[:6]} M | {str(t):<20} | {data_type}\n')

                self.last_tensor_sizes = new_tensor_sizes

            f.write(f"\nAt {where_str:<50}"
                    f" Total Tensor Used Memory:{self.get_tensor_usage():<7.1f}Mb"
                    f" Total Allocated Memory:{self.get_allocate_usage():<7.1f}Mb\n\n")


def torch_info():
    print('Pytorch version\t:', torch.__version__)
    print('CUDA version\t:', torch.version.cuda)
    # 获取CUDNN版本号
    cudnn_version = torch.backends.cudnn.version()
    print("CUDNN version\t:", cudnn_version)
    print('GPU\t\t:', torch.cuda.get_device_name())
    print('GPU count:', torch.cuda.device_count())


def save_batch(batch, filename="batch.pkl"):
    with open(filename, 'wb') as outp:
        pickle.dump(batch, outp, pickle.HIGHEST_PROTOCOL)


@try_except
def model_info(model, pickle_path="batch.pkl", verbose=False, use_pretty=True, input_shape=None, need_profile=True):

    n_p = sum(x.numel() for x in model.parameters())  # number parameters
    n_g = sum(x.numel() for x in model.parameters() if x.requires_grad)  # number gradients
    model_input = None
    if verbose:

        if use_pretty:
            import importlib
            spam_loader = importlib.find_loader('prettytable')
            found = spam_loader is not None
            if found:
                from prettytable import PrettyTable
        else:
            found = False
        
        table = []
        if found:
            if hasattr(model, "_get_name"):
                model_name = model._get_name()
                table.append([' ', '  ',  '   ', model_name, '    ', '     ', '      '])

            table.append(['layer', 'name', 'gradient', 'parameters', 'shape', 'mu', 'sigma'])
        else:
            table.append(f"{'layer':>5} {'name':>40} {'gradient':>9} {'parameters':>12} {'shape':>20} {'mu':>10} {'sigma':>10}")
        
        # for replace model
        # for i, (name, layer) in enumerate(model.named_modules()):
        #     if isinstance(layer, nn.ReLU):  
        #         
        for i, (name, p) in enumerate(model.named_parameters()):
            name = name.replace('module_list.', '')
            if found:
                table.append([i, name, p.requires_grad, p.numel(), list(p.shape),'%.3g' % p.mean().item(), '%.3g' % p.std().item()])
                
            else:
                table.append('%5g %40s %9s %12g %20s %10.3g %10.3g' % (i, name, p.requires_grad, p.numel(), list(p.shape), p.mean(), p.std()))
        if not found:
            for item in table:
                print(item)
        else:
            tab = PrettyTable(table[0], align='r', valign='t')
            tab.add_rows(table[1:])
            print(tab)

    fs = ''
    if need_profile:
        try:
            if pickle_path is not None and os.path.exists(pickle_path):
                with open(pickle_path, 'rb') as inp:
                    model_input = pickle.load(inp)
                    if isinstance(model_input, list):
                        model_input = [data.to(next(model.parameters()).device) for data in model_input]
                    else:
                        model_input.to(next(model.parameters()).device)
            else:
                first_parameter = next(model.parameters())
                if input_shape is None:
                    input_shape = first_parameter.size()
                model_input = torch.zeros(input_shape, device=next(model.parameters()).device).contiguous()  # input
            copy_model = deepcopy(model)
            if isinstance(model_input, list):
                flops = profile(copy_model, inputs=model_input, verbose=False)[0] / 1E9 * 2
            else:
                flops = profile(copy_model, inputs=(model_input,), verbose=False)[0] / 1E9 * 2
            fs = ', %.1f GFLOPs' % (flops)
            copy_model = copy_model.cpu()
            del copy_model
        except Exception as e:
            print(e)
            fs = ''

    print(f"summary: {len(list(model.modules()))} layers, {n_p} parameters, {n_g} gradients{fs}")
    return model_input



from collections import OrderedDict
from turtle import distance
import numpy as np
import cv2
from image_utils import *
import json
import zlib
try:
    import cPickle as pickle
except ModuleNotFoundError:
    import pickle


def _detect_face(detector, img, align=True):
    resp = []

    detected_face = None
    img_region = [0, 0, img.shape[0], img.shape[1]]

    faces = []
    try:
        #faces = detector["face_detector"].detectMultiScale(img, 1.3, 5)
        faces = detector["face_detector"].detectMultiScale(img, 1.1, 10)
    except:
        pass

    if len(faces) > 0:

        for x, y, w, h in faces:
            detected_face = img[int(y):int(y+h), int(x):int(x+w)]

            if align:
                detected_face = align_face(
                    detector["eye_detector"], detected_face)

            img_region = [x, y, w, h]

            resp.append((detected_face, img_region))

    return resp


def align_face(eye_detector, img):

    # eye detector expects gray scale image
    detected_face_gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    #eyes = eye_detector.detectMultiScale(detected_face_gray, 1.3, 5)
    eyes = eye_detector.detectMultiScale(detected_face_gray, 1.1, 10)

    # ----------------------------------------------------------------

    # opencv eye detectin module is not strong. it might find more than 2 eyes!
    # besides, it returns eyes with different order in each call (issue 435)
    # this is an important issue because opencv is the default detector and ssd also uses this
    # find the largest 2 eye. Thanks to @thelostpeace

    eyes = sorted(eyes, key=lambda v: abs(
        (v[0] - v[2]) * (v[1] - v[3])), reverse=True)

    # ----------------------------------------------------------------

    if len(eyes) >= 2:

        # decide left and right eye

        eye_1 = eyes[0]
        eye_2 = eyes[1]

        if eye_1[0] < eye_2[0]:
            left_eye = eye_1
            right_eye = eye_2
        else:
            left_eye = eye_2
            right_eye = eye_1

        # -----------------------
        # find center of eyes
        left_eye = (int(left_eye[0] + (left_eye[2] / 2)),
                    int(left_eye[1] + (left_eye[3] / 2)))
        right_eye = (int(right_eye[0] + (right_eye[2]/2)),
                     int(right_eye[1] + (right_eye[3]/2)))
        img = alignment_procedure(img, left_eye, right_eye)
    return img  # return img anyway


class FaceDetector:

    face_detector = build_model()

    # def build_model(self, detector_backend):
    #     if 'opencv' == detector_backend:
    #         self.face_detector

    @staticmethod
    def detect_face(detector_backend, img, align=True):
        # detector_backend
        obj = _detect_face(FaceDetector.face_detector, img, align)
        # print("---detect faces", len(obj))
        if len(obj) > 0:
            if len(obj) == 1:
                face, region = obj[0]  # discard multiple faces
            else:
                face = None
                region = [0, 0, img.shape[0], img.shape[1]]
        else:  # len(obj) == 0
            face = None
            region = [0, 0, img.shape[0], img.shape[1]]

        return face, region


def detect_face(img, detector_backend='opencv', grayscale=False, enforce_detection=True, align=True):

    img_region = [0, 0, img.shape[0], img.shape[1]]

    # ----------------------------------------------
    # people would like to skip detection and alignment if they already have pre-processed images
    if detector_backend == 'skip':
        return img, img_region

    # ----------------------------------------------

    # detector stored in a global variable in FaceDetector object.
    # this call should be completed very fast because it will return found in memory
    # it will not build face detector model in each call (consider for loops)
    # face_detector = FaceDetector.build_model(detector_backend)

    try:
        # detected_face, img_region = FaceDetector.detect_face(face_detector, detector_backend, img, align)
        detected_face, img_region = FaceDetector.detect_face(
            detector_backend, img, align)
    # if detected face shape is (0, 0) and alignment cannot be performed, this block will be run
    except Exception as e:
        print(e)
        detected_face = None

    if (isinstance(detected_face, np.ndarray)):
        return detected_face, img_region
    else:
        if detected_face == None:
            if enforce_detection != True:
                return img, img_region
            else:
                raise ValueError(
                    "Face could not be detected. Please confirm that the picture is a face photo or consider to set enforce_detection param to False.")


def preprocess_face(
        img, target_size=(224, 224),
        grayscale=False, enforce_detection=True, detector_backend='opencv', return_region=False, align=True, base64_img=False):

    # img might be path, base64 or numpy array. Convert it to numpy whatever it is.
    img = load_image(img, base64_img=base64_img)
    base_img = img.copy()

    img, region = detect_face(img=img, detector_backend=detector_backend,
                              grayscale=grayscale, enforce_detection=enforce_detection, align=align)

    # --------------------------

    if img.shape[0] == 0 or img.shape[1] == 0:
        if enforce_detection == True:
            raise ValueError("Detected face shape is ", img.shape,
                             ". Consider to set enforce_detection argument to False.")
        else:  # restore base image
            img = base_img.copy()

    # --------------------------

    # post-processing
    if grayscale == True:
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    # ---------------------------------------------------
    # resize image to expected shape

    # img = cv2.resize(img, target_size) #resize causes transformation on base image, adding black pixels to resize will not deform the base image

    if img.shape[0] > 0 and img.shape[1] > 0:
        factor_0 = target_size[0] / img.shape[0]
        factor_1 = target_size[1] / img.shape[1]
        factor = min(factor_0, factor_1)

        dsize = (int(img.shape[1] * factor), int(img.shape[0] * factor))
        img = cv2.resize(img, dsize)

        # Then pad the other side to the target size by adding black pixels
        diff_0 = target_size[0] - img.shape[0]
        diff_1 = target_size[1] - img.shape[1]
        if grayscale == False:
            # Put the base image in the middle of the padded image
            img = np.pad(img, ((diff_0 // 2, diff_0 - diff_0 // 2),
                               (diff_1 // 2, diff_1 - diff_1 // 2), (0, 0)), 'constant')
        else:
            img = np.pad(img, ((diff_0 // 2, diff_0 - diff_0 // 2),
                               (diff_1 // 2, diff_1 - diff_1 // 2)), 'constant')

    # ------------------------------------------

    # double check: if target image is not still the same size with target.
    if img.shape[0:2] != target_size:
        img = cv2.resize(img, target_size)

    # ---------------------------------------------------

    # normalizing the image pixels

    img_pixels = img_to_array(img)  # what this line doing? must?
    img_pixels = np.expand_dims(img_pixels, axis=0)
    img_pixels /= 255  # normalize input in [0, 1]

    # ---------------------------------------------------

    if return_region == True:
        return img_pixels, region
    else:
        return img_pixels


class _Layer:
    input_shape = (None, 112, 112, 3)
    output_shape = (None, 1, 128)


class SFaceModel:

    def __init__(self, model_path):

        self.model = cv2.FaceRecognizerSF.create(
            model=model_path,
            config="",
            backend_id=0,
            target_id=0)

        self.layers = [_Layer()]

    def predict(self, image):
        # Preprocess
        # revert the image to original format and preprocess using the model
        input_blob = (image[0] * 255).astype(np.uint8)

        # Forward
        embeddings = self.model.feature(input_blob)

        return embeddings


def represent(
        img_path, model_name='SFaceModel', model=None, enforce_detection=True, detector_backend='opencv', align=True,
        normalization='base', encode=False, base64_img=False):
    embedding = None
    try:
        # decide input shape
        input_shape_x, input_shape_y = find_input_shape(model)

        #detect and align
        img = preprocess_face(img=img_path, target_size=(input_shape_y, input_shape_x),
                              enforce_detection=enforce_detection, detector_backend=detector_backend, align=align, base64_img=base64_img)

        # ---------------------------------
        # custom normalization

        img = normalize_input(img=img, normalization=normalization)
        # represent
        embedding = model.predict(img)[0].tolist()
        # print(embedding)
        if encode:
            content = json.dumps(embedding)
            content = zlib.compress(content.encode('utf-8'))
            content = base64.b64encode(content)
            embedding = str(content, 'utf-8')
    except:
        pass
    return embedding


# https://www.delftstack.com/zh/howto/python/cosine-similarity-between-lists-python/
def partition_arg_topK(matrix, K, axis=0):
    """
    perform topK based on np.argpartition
    :param matrix: to be sorted
    :param K: select and sort the top K items
    :param axis: 0 or 1. dimension to be sorted.
    :return:
    """
    a_part = np.argpartition(matrix, K, axis=axis)
    if axis == 0:
        row_index = np.arange(matrix.shape[1 - axis])
        a_sec_argsort_K = np.argsort(
            matrix[a_part[0:K, :], row_index], axis=axis)
        return a_part[0:K, :][a_sec_argsort_K, row_index]
    else:
        column_index = np.arange(matrix.shape[1 - axis])[:, None]
        a_sec_argsort_K = np.argsort(
            matrix[column_index, a_part[:, 0:K]], axis=axis)
        return a_part[:, 0:K][column_index, a_sec_argsort_K]


def topk(array, k, reverse=True):
    ind = np.argpartition(array, -k)[-k:]
    if reverse:
        newind = np.argsort(array[ind])[::-1][:k]
    else:
        newind = np.argsort(array[ind])
    return ind[newind]


def topk_simlilar(List1, List2, k):
    # print(List1.shape, List2.shape)
    if isinstance(List1, list):
        List1 = np.array(List1)
    if isinstance(List2, list):
        List2 = np.array(List2)
    similarity_scores = np.dot(List1, List2) / (np.linalg.norm(List1, axis=1) * np.linalg.norm(List2))
    distance = 1 - similarity_scores

    threshold = findThreshold('SFace', 'cosine')
    if np.max(distance) > threshold:
        return None

    # print(distance)
    # print(findCosineDistance(List1.tolist()[0], List2.tolist()), findCosineDistance(List2.tolist(), List1.tolist()[0]))
    topk_index = topk(similarity_scores, k, reverse=True)
    return {"topk": topk_index.tolist(), "similarity": similarity_scores[topk_index].tolist()}


def test_cosine_similarity():
    # from numpy import dot
    # from numpy.linalg import norm
    np.random.seed(1)
    # List1 = np.array([4, 47, 8, 3])
    # List2 = np.array([3, 52, 12, 16])
    # result = dot(List1, List2)/(norm(List1)*norm(List2))
    # print(result)
    k = 1

    # List1 = np.array([[4, 47, 8, 3],
    #                   [1, 2000,  100,  2],
    #                   [2, 23,  6,  4],
    #                   [1, 2,  200,  2]])

    List1 = np.random.random((2, 128))
    List2 = np.random.random((1, 128))
    List2 = List2.squeeze()
    similarity_scores = List1.dot(List2) / (np.linalg.norm(List1, axis=1) * np.linalg.norm(List2))
    print(List1.shape, List2.shape)
    print(similarity_scores,  np.dot(List1, List2) / (np.linalg.norm(List1, axis=1) * np.linalg.norm(List2)))
    print(topk(similarity_scores, k, reverse=True))

    # np.array(self.qa_set.questions)[torch.topk(similarity, k)[1]]

    import torch
    import torch.nn.functional as F

    a = torch.FloatTensor(List1.tolist())

    # t2 = [4,54,3,7]
    b = torch.FloatTensor(List2.reshape(1, -1))
    # print(a.shape, b.shape)
    result = F.cosine_similarity(a, b, dim=1, eps=1e-8)
    print(result, torch.topk(result, k)[1])

    # from sklearn.metrics.pairwise import cosine_similarity, cosine_distances
    # A = List1
    # B = List2
    # result = cosine_similarity(A.reshape(1, -1), B.reshape(1, -1))
    # print(result)


def verify_face(image_path, model, embeds, k=5, base64_img=False):
    verify_embed = represent(img_path=image_path, model=model, base64_img=base64_img)
    if verify_embed is None:
        return None

    if isinstance(verify_embed, list):
        verify_embed = np.array(verify_embed)

    # verify_embed = verify_embed.reshape(1, -1)
    if isinstance(embeds, list):
        embeds = np.array(embeds)
    if len(embeds.shape) == 1:
        k = 1
        embeds = embeds.reshape(1, -1)
    else:
        k = min(k, embeds.shape[0])
    # print(verify_embed.shape, embeds.shape, k)
    return topk_simlilar(embeds, verify_embed, k=k)


class ImageEmbeddings:

    def __init__(self) -> None:
        self.order_dict = OrderedDict()
        self.key_index = []
        self.embeds = []

    # add and update
    def add_file(self, file_path, model):
        if not os.path.exists(file_path):
            return False
        filename = os.path.basename(file_path)
        filename = os.path.splitext(filename)[0]
        embed = represent(img_path=file_path, model=model)
        if filename not in self.order_dict:
            self.key_index.append(filename)
        self.order_dict[filename] = embed
        return True

    def remove(self, key):
        if key in self.order_dict:
            del self.order_dict[key]
            self.key_index.remove(key)
            self.generate_embeds()
            return True
        else:
            return False

    def generate_embeds(self):
        self.embeds = []
        for key, val in self.order_dict.items():
            self.embeds.append(val)
        return self.embeds

    def verify_face(self, file_path, model, base64_img=False):
        if not base64_img and os.path.exists(file_path) or base64_img:
            if len(self.embeds) == 0:
                self.generate_embeds()
            ret = verify_face(file_path, model, self.embeds, base64_img=base64_img)
            try:
                ret["files"] = np.array(self.key_index)[ret["topk"]].tolist()
            except Exception as e:
                print(e)
            return ret

    def save(self, filename='embeds.pkl'):
        with open(filename, 'wb') as outp:
            pickle.dump(self, outp, pickle.HIGHEST_PROTOCOL)


def load_embeds(filename='embeds.pkl'):
    try:
        if os.path.exists(filename):
            with open(filename, 'rb') as inp:
                embeds = pickle.load(inp)
                return embeds
    except:
        return None


#: 'WzEsIDIsIDNd'
# decoded_list = base64.b64decode(b64_encoded_list)
# #: '[1, 2, 3]'
# my_list_again = json.loads(decoded_list)
# #: [1, 2, 3]
if __name__ == "__main__":
    model = SFaceModel("/home/faith/.deepface/weights/face_recognition_sface_2021dec.onnx")
    # print(model)
    # embeds = []
    # image_path = "/home/faith/MUSTPHD/deepface/me2.jpg"
    # embeds.append(represent(img_path=image_path, model=model))

    # image_path = "/home/faith/MUSTPHD/deepface/me.jpg"
    # embeds.append(represent(img_path=image_path, model=model))

    # topk_face = verify_face("/home/faith/MUSTPHD/deepface/me2.jpg", model, embeds)
    # print(topk_face)

    # b64_encoded_list = represent(img_path=image_path, model=model, encode=True)
    # print(len(b64_encoded_list))

    embed = ImageEmbeddings()
    embed.add_file("/home/faith/MUSTPHD/deepface/me2.jpg", model)

    # denglun
    # embed.add_file("/home/faith/aicmder/me2.png")

    embed.add_file("/home/faith/MUSTPHD/deepface/me.jpg", model)

    # embed.remove("me2")

    embed.save()
    e = load_embeds()
    print(e.order_dict.keys(), e.key_index)
    # print(e.verify_face("/home/faith/aicmder/dl/IMG_5972.JPG", model))
    # print(e.verify_face("/home/faith/MUSTPHD/deepface/me.jpg", model))

    image_file = "/home/faith/MUSTPHD/deepface/me.jpg"
    with open(image_file,  'rb') as img_f:
        img = img_f.read()
        img_base64 = base64.b64encode(img).decode('utf8')
        print(e.verify_face(img_base64, model, base64_img=True))

    # print(e.verify_face("/home/faith/aicmder/dl/IMG_5029.jpg", model))
    # save(embed)

    # decoded_list = base64.b64decode(b64_encoded_list)
    # print(decoded_list, type(decoded_list))
    # embedding = json.loads(decoded_list.decode())
    # print(embedding)

    # test_cosine_similarity()



import torch
import torch.nn as nn
from torch import Tensor
# https://github.com/iamfaith/attentions
from aicmder.torch.transformer import PositionwiseFeedForward, ActivationFunction, ACTIVATION_FUNCTIONS
import numpy as np
from typing import Optional, Tuple
from torch.nn.init import xavier_uniform_


class DotProductAttention(nn.Module):
    """
    Compute the dot products of the query with all values and apply a softmax function to obtain the weights on the values

    input: batch x num_tree x input_len x dim  16,50,10,100 or 16,50,1000
    w_attn: num_tree x 1     50,1

    input x w_attn = 16,50,1000 x 50,1

    ref: https://discuss.pytorch.org/t/what-is-the-difference-between-register-buffer-and-register-parameter-of-nn-module/32723/8

    """

    def __init__(
            self, atten_dim, d_ff=1024, use_ff=True, activation_fn=ActivationFunction.softmax,
            dropout=0.2):
        super(DotProductAttention, self).__init__()
        self.atten_dim = atten_dim
        data = np.random.uniform(0, 1, (self.atten_dim))
        # self.w_attn = torch.tensor(data, requires_grad=True, dtype=torch.float32)
        # using buffer instead
        self.register_buffer('w_attn', torch.tensor(data, requires_grad=True, dtype=torch.float32))

        self.feed_forward = PositionwiseFeedForward(
            self.atten_dim, d_ff=d_ff, dropout=dropout, activation_fn=ActivationFunction.relu) if use_ff else None
        self.activation_fn = activation_fn

    def forward(self, repr: Tensor) -> Tuple[Tensor, Tensor]:

        batch_size, input_len, hidden_dim = repr.size(0), repr.size(1), repr.size(2)

        # (batch_size, tree_num, dim) x (1, tree_num, 1)
        if self.feed_forward is not None:
            temp_w_attn = self.feed_forward(self.w_attn)
        else:
            temp_w_attn = self.w_attn
        aggregated_repr = torch.einsum('aik,bij->aik', [repr, temp_w_attn.reshape(1, -1, 1)])

        # attention_score = aggregated_repr.reshape([batch_size, -1, hidden_dim])
        attention_score = aggregated_repr

        attention_weights = ACTIVATION_FUNCTIONS[self.activation_fn](attention_score, dim=1)

        weighted_average = torch.multiply(repr, attention_weights)
        weighted_average = torch.sum(weighted_average, dim=1)
        return weighted_average, attention_weights

    def update_dropout(self, dropout):
        self.feed_forward.update_dropout(dropout)


class _DotProductAttention(nn.Module):
    """
    Compute the dot products of the query with all values and apply a softmax function to obtain the weights on the values

    input: batch x num_tree x input_len x dim  16,50,10,100
    w_attn: dim x 1   or   input_len x dim     10,100,1
    """

    def __init__(self, hidden_dim):
        super(DotProductAttention, self).__init__()

    def forward(self, repr: Tensor, w_attn: Tensor) -> Tuple[Tensor, Tensor]:
        assert repr.dim() - w_attn.dim() == 2
        batch_size, input_len, hidden_dim = repr.size(0), repr.size(1), repr.size(2)

        flat_shape = list(w_attn.shape)
        flat_shape.insert(0, -1)
        flat_repr = repr.reshape(flat_shape)

        # print(flat_repr.shape, w_attn.shape)
        # aggregated_repr = torch.einsum('ij,jkl->ikl', w_attn, flat_repr)
        aggregated_repr = torch.matmul(flat_repr, w_attn)
        attention_score = aggregated_repr.reshape([-1, input_len, 1])

        attention_weights = F.softmax(attention_score, dim=1)

        weighted_average = torch.multiply(repr, attention_weights)
        weighted_average = torch.sum(weighted_average, dim=1)
        return weighted_average, attention_weights


class EmbeddingAttentionLayer(nn.Module):
    def __init__(self, dim: int, activation_fn=ActivationFunction.softmax) -> None:
        super().__init__()
        self.dim = dim
        self.activation_fn = activation_fn
        self.attention = torch.randn(1, dim)
        self.attention = nn.Parameter(self.attention)
        # def _reset_module_parameters(self, module):
        # print(self.attention[:20])
        # for p in self.attention:
        #     if p.dim() > 1:
        #         xavier_uniform_(p)
        # print(self.attention[:20])

    def compute_weights(self, embedded: torch.Tensor) -> torch.Tensor:
        unnormalized_weights = embedded.matmul(self.attention.t())
        attention_weights = ACTIVATION_FUNCTIONS[self.activation_fn](unnormalized_weights, dim=1)
        return attention_weights

    def forward(self, embedded: torch.Tensor) -> torch.Tensor:
        attention_weights = self.compute_weights(embedded)
        # print("forward", attention_weights.shape, embedded.shape)
        weighted = torch.matmul(attention_weights.transpose(1, 2), embedded)
        return weighted


# class ModelDevice:
#     def __init__(self, *args, **kwargs):
#         if 'device' in kwargs:
#             self.device = kwargs['device']
#             for k in args:
#                 k = k.to(self.device)

#     def __enter__(self):
#         pass

#     def __exit__(self, exc_type, exc_val, exc_tb):
#         pass


if __name__ == "__main__":

    import numpy as np
    from aicmder.torch import init_seeds
    init_seeds()
    hidden_size = 100
    num_tree = 50
    path_len = 10
    from aicmder.benchmark import model_info

    #                                     N   C         H   W
    memory_bank = torch.randint(0, 2000, (16, num_tree, path_len, hidden_size), dtype=torch.float64)
    dot_attn = DotProductAttention(atten_dim=num_tree, use_ff=False)
    # dot_attn.to(device='cpu')
    model_info(dot_attn)
    print(memory_bank.shape)
    memory_bank = memory_bank.reshape([memory_bank.size(0), memory_bank.size(1), -1]).contiguous()
    # memory_bank.view([memory_bank.size(0), memory_bank.size(1), -1])
    print(memory_bank.shape)
    # print(memory_bank.shape, memory_bank.size())
    out, _ = dot_attn(memory_bank)
    print(out.shape)

    ea = EmbeddingAttentionLayer(hidden_size)
    model_info(ea)

    test_input = torch.randint(0, 2000, (num_tree, path_len, hidden_size), dtype=torch.float32)
    out = ea(test_input)
    print(test_input.shape, out.shape)
    # print(out)

    test_input = torch.randint(0, 2000, (num_tree, path_len + 20, hidden_size), dtype=torch.float32)
    from aicmder.torch import cpu, gpu
    
    def f(ea, test_input):
        ea.to(gpu())
        # test_input.device = gpu()
        test_input.to(gpu())
        print(id(test_input), test_input.device)
    print(id(test_input), test_input.device)
    # test_input = test_input.to(gpu())
    # f(ea, test_input)
    # with ModelDevice(ea, test_input, device=gpu()):
    print(test_input.device, id(test_input)) #, ea.device)
    out = ea(test_input)
    print(test_input.shape, out.shape, out.device, test_input.device)
